# Message box and ui notification msg builder
Files:
 * messagebox.hpp
 * messagebox.cpp
 * uinotificationmsg.hpp

### UiNotificationBuilder
Utility creates translatable json object do be displayed in vDC addon UI. See [Example](#Example)
Class throws exception `UiNotificationBuilderException` when:
 * adding variable when enum value is not mapped to string
 * getting message using `msg` when message is empty
 * variable added to message has no value

Message level:
 * Success
 * Info
 * Warning
 * Error

### Example:
```
   // create builder, map enum variables to string values
   vdcutils::UiNotificationBuilder builder;
   // build message like pushing data to stream.
   // Use UiNotificationTranslatable to put tranlatable string into the message
   // Use UiNotificationUntranslatable to put untranlatable string into the message
   builder << vdcutils::UiNotificationTranslatable{"device_type", "Washing Machine"}
           << " model "
           << vdcutils::UiNotificationUntranslatable{"device_model", "XYZ-1234"}
           << " is not supported";
   auto j1 = builder.msg("id 1");

   // message without variables
   builder << "Device is ready";
   auto j2 = builder.msg("id 2");


   // create message box, 5 is the buffer size
   vdcutils::MessageBox mb(5);
   // add messages to message box
   mb.addMessage(vdcutils::MessageLevel::Warning, j1);
   mb.addMessage(vdcutils::MessageLevel::Info, j2);
   // get message box json
   auto json = mb.toJson();
   LOG(LOG_NOTICE, ">>>> %s", json->stringValue().c_str());
```

Output:
```
{
   "messages":[
      {
         "timestamp":1557739926,
         "level":"warning",
         "message":{
            "id":"id 1",
            "txt":"${device_type} model ${device_model} is not supported",
            "translatables":{
               "device_type":"Washing Machine"
            },
            "untranslatables":{
               "device_model":"XYZ-1234"
            }
         }
      },
      {
         "timestamp":1557739926,
         "level":"info",
         "message":{
            "id":"id 2",
            "txt":"Device is ready"
         }
      }
   ]
}

```


# Authorization Handler (authorization.hpp)

Authorization Handler is a wrapper class to handle authorization process, like oAuth2.
Class takes care about authorization data storage and refreshing access data.
As a developer you have to implement AuthorizationData class which implements IAuthorizationData interface.
AuthorizationData class is encapsulated in AuthorizationHandler.
AuthorizationHandler will refresh token ahead of time, if AuthorizationData supports expiration time.
If Refresh Token procedure fails, retry is trigerred after 1, 2, 3, 4 and 5 minutes. Then retry is repeated after 5 minutes.


## IAuthorizationData interface
   * serialize() - serialize auth data to p44::JsonObjectPtr.
                   Returned JSON data is stored in db.
                   Return nullptr if does not succeeded, do not throw.
   * deserialize(p44::JsonObjectPtr aJson) - deserizalize JSON loaded from db to auth data.
                                             Do not throw.
   * parseAuthData(p44::JsonObjectPtr aAuthPayload) - parse authorization payload received from provider.
                                                      Throw vdcutils::AuthorizationException if has failed.
   *  getExpireTime() - epoch timestamp in seconds, when authorization expires.
                        Returns std::optional<std::time_t>.
                        If expiration time not supported, return std::nullopt.
   * refreshAccessToken(p44::StatusCB aCompletedCB) - implementation of refreshing auth data procedure.
                                                      Call callback with appriopriate p44::ErrorPtr.
                                                      Do not throw.

## Authorization Handler
   * parseAuthData(p44::JsonObjectPtr aAuthPayload) - parse authorization payload received from provider.
                                                      Calls IAuthorizationData::parseAuthData.
   * refreshAccessToken(p44::StatusCB aCompletedCB) - refresh token on demand.

### Example:

AuthorizationData Implementation
(note: this is example, demo, and yes, it is imperfect implementation)

```
struct AuthData : public vdcutils::IAuthorizationData
  {
    std::string accessToken;
    std::string refreshToken;
    std::string userEmail;
    std::time_t expireTime;
    string clientId;
    string clientSecret;

    vdcutils::HttpClient httpClient;

    AuthData(){
      httpClient.isMemberVariable();
    }

    virtual p44::JsonObjectPtr serialize() override
    {
      auto json = p44::JsonObject::newObj();
      json->add("access_token", p44::JsonObject::newString(accessToken));
      json->add("refresh_token", p44::JsonObject::newString(refreshToken));
      json->add("client_id", p44::JsonObject::newString(clientId));
      json->add("client_secret", p44::JsonObject::newString(clientSecret));
      json->add("expires_in", p44::JsonObject::newInt32(expireTime));
      return json;
    }

    virtual void deserialize(p44::JsonObjectPtr aJson) override
    {
      if (!aJson) {
         return;
      }

      accessToken = aJson->get("access_token")->stringValue();
      refreshToken = aJson->get("refresh_token")->stringValue();
      clientId = aJson->get("client_id")->stringValue();
      clientSecret = aJson->get("client_secret")->stringValue();
      expireTime = aJson->get("expires_in")->int32Value();
    }

    virtual void parseAuthData(p44::JsonObjectPtr aAuthPayload) override
    {
      deserialize(aAuthPayload);
      auto now = (p44::MainLoop::unixtime() / p44::Second);
      expireTime += now;
    }

    virtual std::optional<std::time_t> getExpireTime()
    {
      return  (expireTime > (p44::MainLoop::unixtime() / p44::Second)) ?
              std::make_optional(expireTime)
               : std::nullopt;
    }

    virtual void refreshAccessToken(p44::StatusCB aCompletedCB) override
    {
      requestRefreshToken(aCompletedCB);
    }

    void requestRefreshToken(p44::StatusCB aCompletedCB)
    {
      static constexpr auto AUTHENTICATE_URL = "https://api.netatmo.com/oauth2/token";
      stringstream requestBody;
      requestBody
        << "grant_type=refresh_token"
        << "&refresh_token=" << refreshToken
        << "&client_id="     << clientId
        << "&client_secret=" << clientSecret;

      httpClient.getApi().httpRequest(
        AUTHENTICATE_URL,
        [=] (auto aResponse, auto aError) {
          if (p44::Error::ok(aError)) {
            gotRefreshToken(p44::JsonObject::objFromText(aResponse.c_str()));
          }
          aCompletedCB(p44::Error::ok());
        },
        "POST",
        requestBody.str().c_str(),
        "application/x-www-form-urlencoded;charset=UTF-8");
    }


    void gotRefreshToken(p44::JsonObjectPtr aJson)
    {
      if (!aJson) {
        return;
      }

      accessToken = aJson->get("access_token")->stringValue();
      refreshToken = aJson->get("refresh_token")->stringValue();
      auto now = (p44::MainLoop::unixtime() / p44::Second);
      expireTime = aJson->get("expires_in")->int32Value() + now;
    }

  };

```

AuthorizationHandler usage
```
   std::unique_ptr<vdcutils::AuthorizationHandler<AuthData>> mAuthHandler;

   mAuthHandler = std::make_unique<vdcutils::AuthorizationHandler<AuthData>>(getStorage(), dSUID.getString());


  if (aMethod=="authenticate") {

    string authData;
    respErr = checkStringParam(aParams, "authData", authData);

    if (auto jsonAuthData = JsonObject::objFromText(authData.c_str())) {
      mAuthHandler->parseAuthData(jsonAuthData)
    }
  }

   auto op = vdcutils::make_intrusive<Operation>(
      aQuery,
      httpClient,
      accessData.token,
      apiQueryCB,
      [=](StatusCallback aCB){ mAuthHandler->refreshAccessToken(aCB); }
    );


```