/*
 *  Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of vdc-utils
 */

#include "applicationvdc.hpp"

const string DEFAULT_PBUF_VDSMSERVICE = "8340";
const string VDSM_PORT_OPTION_DESCRIPTION = "port;port number/service name for vdSM to connect to (default:"  + DEFAULT_PBUF_VDSMSERVICE + ")";
const string DEFAULT_DBDIR = "/tmp";
const string DB_DIR_OPTION_DESCRIPTION = "dirpath;set SQLite DB directory (default = " + DEFAULT_DBDIR + ")";
const int DEFAULT_LOGLEVEL = LOG_NOTICE;

using namespace p44;

ApplicationVdc::ApplicationVdc(VdcHostPtr aVdcHost) :
  vdcHost(aVdcHost)
{

}

boost::optional<string> ApplicationVdc::getStringOption(const string& aOptionName)
{
  string s;
  if (getStringOption(aOptionName.c_str(), s)) {
    return s;
  }
  return boost::none;
}

void ApplicationVdc::parseArgs(int argc, char **argv)
{
  const CmdLineOptionDescriptor ops[] = {
      { 0  , "dsuid",         true,   "dSUID;set dSUID for this vDC host (usually UUIDv1 generated on the host)" },
      { 0  , "ifnameformac",  true,   "network if;set network interface to get MAC address from" },
      { 0  , "ifnameforconn", true,   "network if;set network interface to get IP from and check for connectivity" },
      { 0  , "productname",   true,   "name;set product name for this vdc host and its vdcs" },
      { 0  , "productversion",true,   "version;set version string for this vdc host and its vdcs" },
      { 0  , "deviceid",      true,   "device id;a string that may identify the device to the end user, e.g. a serial number" },
      { 0  , "description",   true,   "description(template);used in service announcement, can contain %V,%M,%N,%S to insert vendor/model/name/serial. "
                                       "When not set or empty, defaults to built-in standard description." },
      { 0  , "vdcdescription",true,   "vdcmodelname(template);can contain %V,%M,%m,%S to insert vendor/productname/modelsuffix/serial. "
                                      "When not set or empty, defaults to built-in standard description." },
      { 'C', "vdsmport",      true,  VDSM_PORT_OPTION_DESCRIPTION.c_str() },
      { 'i', "vdsmnonlocal",  false, "allow vdSM connections from non-local clients" },
      { 'w', "startupdelay",  true,  "seconds;delay startup" },
      { 'l', "loglevel",      true,  "level;set max level of log message detail to show on stdout" },
      { 0  , "errlevel",      true,  "level;set max level for log messages to go to stderr as well" },
      { 0  , "dontlogerrors", false, "don't duplicate error messages (see --errlevel) on stdout" },
      { 0  , "deltatstamps",  false, "show timestamp delta between log lines" },
      { 's', "sqlitedir",     true,  DB_DIR_OPTION_DESCRIPTION.c_str() },
      { 0  , "icondir",       true,  "icon directory;specifiy path to directory containing device icons" },
      { 0  , "configdir",     true,  "dirpath;set directory for config files (defaults to sqlitedir)" },
      { 0  , "defaultvdcname",true,  "name;default name for vdc(s), only visible when user has not set a name" },
#if !DISABLE_DISCOVERY
      { 0  , "noauto",        false, "prevent auto-connection to this vdc host" },
      { 0  , "nodiscovery",   false, "completely disable discovery (no publishing of services)" },
      { 0  , "hostname",      true,  "hostname;host name to use to publish this vdc host" },
#endif
#if P44_JSON_API_SUPPORT
      { 'W', "cfgapiport",    true,  "port;server port number for web configuration JSON API (default=none)" },
      { 0  , "cfgapinonlocal",false, "allow web configuration JSON API from non-local clients" },
#endif
#if SSL_ENABLED
      { 0  , "sslcafile",     true,  "SSL CAFile, set ssl certificate file (The default CA certificates file is called \"cert.pem\" in the default OpenSSL directory.)" },
  #if SSL_SWITCH
        { 0  , "insecure",         false, "disable ssl certificate verification" },
  #endif
#endif
      { 'h', "help",          false, "show this text" },
      { 0, NULL } // list terminator
  };

  Options options = Options(ops, ops + sizeof(ops) / sizeof(CmdLineOptionDescriptor));
  Options vdcOpts = vdcOptions();
  options.insert(options.begin(), vdcOpts.begin(), vdcOpts.end());

  const char *usageText = "Usage: %1$s [options]\n";

  // parse the command line, exits when syntax errors occur
  setCommandDescriptors(usageText, options.data());
  parseCommandLine(argc, argv);
}


string ApplicationVdc::getSslCertConfig()
{
#if SSL_ENABLED
  #if SSL_SWITCH
    if (getOption("insecure")) {
      // disable ssl verification
      return "";
    }
  #endif
  boost::optional<string> sslCAFile = getStringOption("sslcafile");
  if (sslCAFile) {
    // set ca file
    return "=" + *sslCAFile;
  }
#endif

  // set default ca path (/etc/ssl/certs/)
  return "*";
}


int ApplicationVdc::main()
{
  // build objects only if not terminated early
  if (!isTerminated()) {

    // log level?
    int loglevel = DEFAULT_LOGLEVEL;
    getIntOption("loglevel", loglevel);
    SETLOGLEVEL(loglevel);
    int errlevel = LOG_ERR; // testing by default only reports to stdout
    getIntOption("errlevel", errlevel);
    SETERRLEVEL(errlevel, !getOption("dontlogerrors"));
    SETDELTATIME(getOption("deltatstamps"));

    // startup delay?
    int startupDelay = 0; // no delay
    getIntOption("startupdelay", startupDelay);

    // before starting anything, delay
    if (startupDelay>0) {
      LOG(LOG_NOTICE, "Delaying startup by %d seconds (-w command line option)", startupDelay);
      sleep(startupDelay);
    }

    // Init the device container root object
    // - set DB dir
    boost::optional<string> dbdir = getStringOption("sqlitedir");
    auto dbdirCstr = dbdir ? dbdir->c_str() : DEFAULT_DBDIR.c_str();
    vdcHost->setPersistentDataDir(dbdirCstr);

    // - set conf dir
    boost::optional<string> confdir = getStringOption("configdir");
    vdcHost->setConfigDir(confdir ? confdir->c_str() : dbdirCstr);

    // - set icon directory
    boost::optional<string> icondir = getStringOption("icondir");
    vdcHost->setIconDir(icondir ? icondir->c_str() : NULL);
    string s;

    // - set dSUID mode
    DsUidPtr externalDsUid;
    if (getStringOption("dsuid", s)) {
      externalDsUid = DsUidPtr(new DsUid(s));
    }
    boost::optional<string> macif = getStringOption("ifnameformac");
    vdcHost->setIdMode(externalDsUid, macif ? *macif : "");

    // - network interface
    if (getStringOption("ifnameforconn", s)) {
      vdcHost->setNetworkIf(s);
    }

    // - set product name and version
    if (getStringOption("productname", s)) {
      vdcHost->setProductName(s);
    }
    // - set product version
    if (getStringOption("productversion", s)) {
      vdcHost->setProductVersion(s);
    }
    // - set product device id (e.g. serial)
    if (getStringOption("deviceid", s)) {
      vdcHost->setDeviceHardwareId(s);
    }
    // - set description (template)
    if (getStringOption("description", s)) {
      vdcHost->setDescriptionTemplate(s);
    }
    // - set vdc modelName (template)
    if (getStringOption("vdcdescription", s)) {
      vdcHost->setVdcModelNameTemplate(s);
    }


    // - set API
    vdcHost->vdcApiServer = VdcApiServerPtr(new VdcPbufApiServer());
    // set up server for vdSM to connect to
    boost::optional<string> vdcapiservice = getStringOption("vdsmport");
    vdcHost->vdcApiServer->setConnectionParams(NULL,
                                               vdcapiservice ? vdcapiservice->c_str() : DEFAULT_PBUF_VDSMSERVICE.c_str(),
                                               SOCK_STREAM,
                                               AF_INET);
    vdcHost->vdcApiServer->setAllowNonlocalConnections(getOption("vdsmnonlocal"));

    // Create class containers

    // - first, prepare (make sure dSUID is available)
    try {
      vdcHost->prepareForVdcs(false);
    }
    catch(const sqlite3pp::database_error& ) {
      LOG(LOG_ERR, "Cannot initialize sqlite db");
      return -1;
    }

    Vdcs vdcs = createVdcs();
    for (Vdcs::iterator it = vdcs.begin(); it != vdcs.end(); ++it) {
      (*it)->addVdcToVdcHost();
    }

    // set default name (if any) for instantiated vdcs
    string defaultname;
    if (getStringOption("defaultvdcname", defaultname)) {
      for (VdcMap::iterator pos=vdcHost->vdcs.begin(); pos!=vdcHost->vdcs.end(); ++pos) {
        (pos->second)->initializeName(defaultname);
      }
    }

  } // if !terminated
  // app now ready to run (or cleanup when already terminated)
  return run();
}


void ApplicationVdc::initialize()
{
  // - initialize the device container
  vdcHost->initialize(boost::bind(&ApplicationVdc::initialized, this, _1), false); // no factory reset
}


void ApplicationVdc::initialized(ErrorPtr aError)
{
  if (!Error::isOK(aError)) {
    // cannot initialize, this is a fatal error
    // exit in 15 seconds
    LOG(LOG_ALERT, "****** Fatal error - vdc host initialisation failed: %s", aError->description().c_str());
    terminateTicket.executeOnce(boost::bind(&ApplicationVdc::terminateAppWith, this, aError), 15*Second);
    return;
  }
  else {
    // Initialized ok
#if !DISABLE_DISCOVERY
    // - initialize discovery
    initDiscovery();
#endif
    // - start running normally
    vdcHost->startRunning();
    // - collect devices
    collectDevices(rescanmode_normal);
  }
}


void ApplicationVdc::collectDevices(RescanMode aRescanMode)
{
  // initiate device collection
  vdcHost->collectDevices(boost::bind(&ApplicationVdc::devicesCollected, this, _1), aRescanMode);
}


void ApplicationVdc::devicesCollected(ErrorPtr aError)
{
  if (Error::isOK(aError)) {
    DBGLOG(LOG_INFO, "%s", vdcHost->description().c_str());
  }
  else {
    LOG(LOG_ERR, "****** Error - collecting devices failed: %s", aError->description().c_str());
  }
}


#if !DISABLE_DISCOVERY

void ApplicationVdc::initDiscovery()
{
  // get discovery params
  // - host name
  string hostname;
  if (!getStringOption("hostname", hostname)) {
    // none specified, create default
    hostname = vdcHost->modelName() + "-" + vdcHost->getDsUid().getString();
  }
  // start the basic service
  ErrorPtr err = DiscoveryManager::sharedDiscoveryManager().start(hostname.c_str());
  if (Error::isOK(err)) {
    // start DS advertising if not disabled
    if (!getOption("nodiscovery")) {
      // started ok, set discovery params
      int sshPort = 0;
      getIntOption("sshport", sshPort);
      // start discovery manager
      DiscoveryManager::sharedDiscoveryManager().advertiseDS(
          vdcHost,
          getOption("noauto"),
          0, // no web port advertised
          0, // no ssh port advertised
          DsUidPtr(),
          0,
          false,
          NULL,
          false
      );
    }
  }
  else {
    LOG(LOG_ERR, "**** Cannot start discovery manager: %s", err->description().c_str());
  }
}

#endif // !DISABLE_DISCOVERY

