/*
 *  Copyright (c) 2019 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of vdc-utils
 */

#pragma once

#include <ctime>
#include <type_traits>
#include "exception.hpp"
#include "persistentstorage/persistentstorage.hpp"
#include "persistentstorage/persistentstoragefactory.hpp"
#include "p44utils/jsonobject.hpp"

namespace vdcutils {

using AuthorizationException = vdcutils::Exception;

class IAuthorizationData
{
  public:
  virtual ~IAuthorizationData()=default;
  /* serialize object to json for db blob */
  virtual p44::JsonObjectPtr serialize() = 0;
  /* load json blob data from db to object */
  virtual void deserialize(p44::JsonObjectPtr aJson) = 0;
  /* parse providers authorization payload */
  virtual void parseAuthData(p44::JsonObjectPtr aAuthPayload) = 0;
  /* epoch timestamp in seconds of data expiration */
  virtual std::optional<std::time_t> getExpireTime() { return std::nullopt; }
  /* refresh access procedure */
  virtual void refreshAccessToken(p44::StatusCB aCompletedCB) = 0;
};

class RefreshTokenCounter
{
  unsigned int counter;
  static constexpr unsigned int REFRESH_TOKEN_INTERVAL_MAX = 5;
public:
  RefreshTokenCounter() :
    counter(0) {}
  void reset() { counter = 0; }
  p44::MLMicroSeconds getInterval()
  {
    if (counter < REFRESH_TOKEN_INTERVAL_MAX) {
      ++counter;
    }
    return counter * p44::Minute;
  }
};

template<typename AuthData,
         typename = std::enable_if_t<
           std::is_base_of_v<IAuthorizationData, AuthData>>>
class AuthorizationHandler;


template<typename AuthData>
class AuthorizationHandler<AuthData>
{
  static constexpr auto TIME_BEFORE_EXPIRES {15*p44::Minute};
  AuthData mData;
  std::unique_ptr<vdcutils::IPersistentStorageWithRowId> mStorage;
  p44::MLTicket mRefreshTokenTicket;
  RefreshTokenCounter mRefreshTokenCounter;

  public:
  template<typename... AuthDataArgs>
  AuthorizationHandler(IPersistentStorageFactory& aStorage,
                       const std::string& aRowId,
                       AuthDataArgs&&... aAuthDataArgs) :
    mData(std::forward<AuthDataArgs>(aAuthDataArgs)...)
  {
    createPersistentRecord(aStorage);
    insertTable(aStorage, aRowId);
  }

  void parseAuthData(p44::JsonObjectPtr aAuthPayload);
  void refreshAccessToken(p44::StatusCB aCompletedCB);

  private:
  void createPersistentRecord(IPersistentStorageFactory& aStorage);
  void insertTable(IPersistentStorageFactory& aStorage, const std::string& aRowId);
  void scheduleRefreshToken();
};


template<typename T>
void AuthorizationHandler<T>::createPersistentRecord(IPersistentStorageFactory& aStorage)
{
  aStorage.addRecord(
    PersistentRecordCB<std::string>({
      "blob",
      [=](const std::string& aValue){
        mData.deserialize(p44::JsonObject::objFromText(aValue.c_str()));
      },
      [=](){
        auto jsonAuth = mData.serialize();
        return jsonAuth ? jsonAuth->stringValue() : std::string();
      }
    }));
}


template<typename T>
void AuthorizationHandler<T>::insertTable(IPersistentStorageFactory& aStorage, const std::string& aRowId)
{
  mStorage = aStorage.createStorageWithRowId(aRowId, "Authorization");
  mStorage->load();
  scheduleRefreshToken();
}


template<typename T>
void AuthorizationHandler<T>::parseAuthData(p44::JsonObjectPtr aAuthPayload)
{
  if (!aAuthPayload) {
    throw AuthorizationException("Authorization payload is empty");
  }

  try {
    mData.parseAuthData(aAuthPayload);
    mStorage->save();
    scheduleRefreshToken();
  } catch (AuthorizationException& e) {
    LOG(LOG_ERR, "[AuthHandler] Cannot parse auth data. Exception: '%s'", e.what());
    throw e;
  }
}


template<typename T>
void AuthorizationHandler<T>::scheduleRefreshToken()
{
  auto expireTime = mData.getExpireTime();
  if (!expireTime) {
    LOG(LOG_NOTICE, "[AuthHandler] Cancel pending refresh token");
    mRefreshTokenTicket.cancel();
    return;
  }

  auto mlTimestamp = p44::MainLoop::unixTimeToMainLoopTime(expireTime.value() * p44::Second);
  // refresh ahead of time
  mlTimestamp = mlTimestamp - TIME_BEFORE_EXPIRES;
  LOG(LOG_NOTICE, "[AuthHandler] Schedule refresh token 15 minutes before %s (%llu)",
                  std::ctime(&expireTime.value()), expireTime.value());

  mRefreshTokenTicket.executeOnceAt(
    [=](auto...){ refreshAccessToken([](auto...){}); },
    mlTimestamp, /* execution time */
    p44::Minute /* tolerance */
  );

}

template<typename T>
void AuthorizationHandler<T>::refreshAccessToken(p44::StatusCB aCompletedCB)
{
  mData.refreshAccessToken([=](auto aError){
    if (!p44::Error::isOK(aError)) {
      LOG(LOG_ERR, "[AuthHandler] Refresh token has failed ('%s')", aError->description().c_str());
      auto retryDelay = mRefreshTokenCounter.getInterval();
      LOG(LOG_NOTICE, "[AuthHandler] Retry to refresh token in '%llu' minutes", retryDelay/p44::Minute);
      mRefreshTokenTicket.executeOnce(
          [=](auto...){ refreshAccessToken([](auto...){}); },
          retryDelay
      );
      if (aCompletedCB) aCompletedCB(aError);
      return;
    }

    mStorage->save();
    mRefreshTokenCounter.reset();
    scheduleRefreshToken();
    if (aCompletedCB) aCompletedCB(aError);
  });
}

} // namespace vdcutils
