/*
 *  Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of vdc-utils
 */
#pragma once

#if __cplusplus >= 201402L

#include <string>
#include "p44utils/error.hpp"
#include "p44utils/operationqueue.hpp"
#include "p44utils/httpcomm.hpp"

namespace vdcutils {

  enum class AccountStatus {
    disconnected,
    offline,
    connected,
    unknown
  };

  const int HTTP_OK_ERR_CODE = 200;
  const int HTTP_CREATED = 201;
  const int HTTP_BAD_REQUEST = 400;
  const int HTTP_UNAUTHORIZED_ERR_CODE = 401;
  const int HTTP_FORBIDDEN_ERR_CODE = 403;
  const int HTTP_NOT_FOUND_ERR_CODE = 404;
  const int HTTP_INTERNAL_SERVER_ERROR = 500;
  const int HTTP_SERVICE_UNAVAILABLE_ERR_CODE = 503;

  using IsAuthErrorCB = std::function<bool(p44::ErrorPtr)>;

  std::string toString(AccountStatus aValue);
  bool isHttpCommError(p44::ErrorPtr aError);
  bool isServerError(p44::ErrorPtr aError);
  bool isAuthorizationError(p44::ErrorPtr aError);
  AccountStatus updateAccountStatus(p44::ErrorPtr aError, IsAuthErrorCB aIsAuthError=isAuthorizationError);



template <typename Comm>
class Client : public p44::OperationQueue
{
    using inherited = p44::OperationQueue;

    Comm commApi;
    AccountStatus accountStatus = AccountStatus::unknown;
    IsAuthErrorCB isAuthErrorCB;

  public:
    Client(IsAuthErrorCB aAuthErrorCB = [](p44::ErrorPtr aError){ return !p44::Error::isOK(aError) &&
                                                                    (aError->getErrorCode() == HTTP_UNAUTHORIZED_ERR_CODE || aError->getErrorCode() == HTTP_FORBIDDEN_ERR_CODE); }) :
      inherited(p44::MainLoop::currentMainLoop()),
      commApi(p44::MainLoop::currentMainLoop()),
      isAuthErrorCB(aAuthErrorCB)
    {
      commApi.isMemberVariable();
    }

    virtual ~Client() {}

    Comm& getApi() { return commApi; }

    void enqueueAndProcessOperation(p44::OperationPtr aOperation)
    {
      queueOperation(aOperation);
      processOperations();
    }

    // queue a new operation
    void queueFrontOperation(p44::OperationPtr aOperation)
    {
      operationQueue.push_front(aOperation);
    }

    void enqueueFrontAndProcessOperation(p44::OperationPtr aOperation)
    {
      queueFrontOperation(aOperation);
      processOperations();
    }

    bool isAuthError(p44::ErrorPtr aError) { return isAuthErrorCB(aError); }

    void updateAccountStatus(p44::ErrorPtr aError)
    {
      accountStatus = vdcutils::updateAccountStatus(aError, isAuthErrorCB);
    }

    AccountStatus getAccountStatus() { return accountStatus; }

};

} // namespace vdcutils

#endif

