/*
 *  Copyright (c) 2019 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of vdc-utils
 */

#pragma once

#include <type_traits>
#include "vdc/vdcwithproperties.hpp"
#include "persistentstorage/persistentstorage.hpp"
#include "persistentstorage/persistentstoragefactory.hpp"
#include "p44utils/jsonobject.hpp"


namespace vdcutils {

class IDiscoveredDevice
{
  public:
  virtual ~IDiscoveredDevice()=default;
  virtual p44::JsonObjectPtr serialize() = 0;
  virtual bool deserialize(p44::JsonObjectPtr aJson) = 0;
  virtual p44::JsonObjectPtr getPropertyNotation() = 0;
  virtual std::string getId() const = 0;
  bool operator==(const IDiscoveredDevice& aDevice) {
    return this->getId() == aDevice.getId();
  }
  bool operator==(const std::string& aId) {
    return this->getId() == aId;
  }
};

template<typename T,
         typename = std::enable_if_t<
           std::is_base_of_v<IDiscoveredDevice, T>
           && std::is_default_constructible_v<T>>>
class IDiscoveredDevices;

template<typename T>
class IDiscoveredDevices<T>
{
  public:
  using Devices_t = std::vector<T>;
  using RemoveIfPredicate_t = std::function<bool (const T&)>;
  virtual ~IDiscoveredDevices() = default;

  virtual void addDevice(const T& aNewDevice) = 0;
  virtual void addOrUpdateDevice(const T& aNewDevice) = 0;
  virtual void removeDevice(const std::string& aId) = 0;
  virtual std::optional<T> getDevice(const std::string& aId) const = 0;
  virtual const Devices_t getDevices() const = 0;
  virtual void removeIf(RemoveIfPredicate_t aPredicate) = 0;
};


template<typename T>
class DiscoveredDevices : public IDiscoveredDevices<T>
{
  using StringList_t = std::list<std::string>;
  using Devices_t = std::vector<T>;
  using RemoveIfPredicate_t = std::function<bool (const T&)>;
  std::unique_ptr<vdcutils::IPersistentStorageWithRowId> mStorage;
  Devices_t mDevices;

  public:
  template<typename Dev>
  DiscoveredDevices(
      VdcWithProperties<Dev>& aVdc,
      const std::string& aPropertyName,
      const std::string& aTableName,
      const std::string& aRecordName)
  {
    aVdc.addProperty(Property<string>{
        aPropertyName,
        [&](){ return this->containerToJsonString(); },
        [](auto...){}
      });

    createPersistentRecord(aVdc.getStorage(),
                           aVdc.getDsUid().getString(),
                           aTableName,
                           aRecordName);
  }

  private:
  void createPersistentRecord(IPersistentStorageFactory& aStorage,
                              const std::string& aRowId,
                              const std::string& aTableName,
                              const std::string& aRecordName) {
    aStorage.addRecord(
        PersistentRecordCB<StringList_t>({
          aRecordName,
          [=](const StringList_t& aValue){
            this->deserializeFromStorage(aValue);
          },
          [=](){
            return this->serializeToStorage();
          }
    }));
    mStorage = aStorage.createStorageWithRowId(aRowId, aTableName);
    mStorage->load();
  }

  void deserializeFromStorage(const StringList_t& aDevices) {
    mDevices.clear();
    for (auto& device : aDevices) {
      auto json = p44::JsonObject::objFromText(device.c_str());
      if (!json) {
        LOG(LOG_ERR, "Malformed JSON: '%s'", device.c_str());
        return;
      }
      LOG(LOG_NOTICE, "Adding device from storage: '%s", json->stringValue().c_str());
      if (T dev;
          dev.deserialize(json)) {
        mDevices.push_back(dev);
      }
    }
  }

  StringList_t serializeToStorage() {
    StringList_t devices;
    for (auto& device : mDevices) {
      auto json = device.serialize();
      if (json) {
        devices.emplace_back(json->stringValue());
      }
    }
    return devices;
  }

  std::string containerToJsonString() {
    auto jsonArray = p44::JsonObject::newArray();
    for (auto& device : mDevices) {
      auto json = device.getPropertyNotation();
      if (json) {
        jsonArray->arrayAppend(json);
      }
    }
    return jsonArray->stringValue();
  }

  public:

  void addDevice(const T& aNewDevice) {
    auto duplicated = std::any_of(mDevices.begin(), mDevices.end(),
        [=](auto& aDevice) { return aDevice == aNewDevice; });
    if (duplicated) {
      LOG(LOG_DEBUG, "Device with id '%s' already exist", aNewDevice.getId().c_str());
      return;
    }
    mDevices.emplace_back(std::move(aNewDevice));
    mStorage->save();
  }

  void addOrUpdateDevice(const T& aNewDevice) {
    auto it = std::find(mDevices.begin(), mDevices.end(), aNewDevice);
    if (it != mDevices.end()) {
      mDevices.erase(it);
    }
    LOG(LOG_DEBUG, "Updating/adding device with id '%s'", aNewDevice.getId().c_str());
    mDevices.emplace_back(std::move(aNewDevice));
    mStorage->save();
  }

  void removeDevice(const std::string& aId) {
    auto it = std::find_if(mDevices.begin(), mDevices.end(),
        [=](auto aDevice) { return aDevice == aId; });
    if (it != mDevices.end()) {
      LOG(LOG_DEBUG, "Removing device with id '%s'", it->getId().c_str());
      mDevices.erase(it);
      mStorage->save();
    }
  }

  std::optional<T> getDevice(const std::string& aId) const {
    auto it = std::find_if(mDevices.begin(), mDevices.end(),
        [=](auto aDevice) { return aDevice == aId; });
    if (it == mDevices.end()) {
      return std::nullopt;
    }
    return *it;
  }

  const Devices_t getDevices() const {
    return mDevices;
  }

  void removeIf(RemoveIfPredicate_t aPredicate) {
    mDevices.erase(
        std::remove_if(mDevices.begin(), mDevices.end(), aPredicate),
        mDevices.end()
    );
  }

};


} // namespace vdcutils
