/*
 *  Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of vdc-utils
 */

#include <boost/algorithm/string.hpp>

#include "dsvdchost.hpp"
#include "vdc/vdc.hpp"

namespace p44 {

// MARK: ===== DB and initialisation


DsVdcHost::DsVdcHost() : vdcCount(1), vdcConfiguration("1;")
{
  setProductName(PACKAGE_NAME);
  setProductVersion(PACKAGE_VERSION);
  setDescriptionTemplate(DSVDC_DESCRIPTION_TEMPLATE);
  setVdcModelNameTemplate(DSVDC_VDCMODELNAME_TEMPLATE);
}

// data field definitions
static const size_t numFields = 2;

size_t DsVdcHost::numFieldDefs()
{
  return inheritedParams::numFieldDefs()+numFields;
}

const FieldDefinition *DsVdcHost::getFieldDef(size_t aIndex)
{
  static const FieldDefinition dataDefs[numFields] = {
    { "vdcCount", SQLITE_INTEGER },
    { "vdcConfiguration", SQLITE_TEXT }
  };
  if (aIndex<inheritedParams::numFieldDefs())
    return inheritedParams::getFieldDef(aIndex);
  aIndex -= inheritedParams::numFieldDefs();
  if (aIndex<numFields)
    return &dataDefs[aIndex];
  return NULL;
}

/// load values from passed row
void DsVdcHost::loadFromRow(sqlite3pp::query::iterator &aRow, int &aIndex, uint64_t *aCommonFlagsP)
{
  inheritedParams::loadFromRow(aRow, aIndex, aCommonFlagsP);
  // get the field value
  vdcCount = aRow->get<int>(aIndex++);
  aRow->getIfNotNull(aIndex++, vdcConfiguration);
}

// bind values to passed statement
void DsVdcHost::bindToStatement(sqlite3pp::statement &aStatement, int &aIndex, const char *aParentIdentifier, uint64_t aCommonFlags)
{
  inheritedParams::bindToStatement(aStatement, aIndex, aParentIdentifier, aCommonFlags);
  // bind the fields
  aStatement.bind(aIndex++, vdcCount);
  aStatement.bind(aIndex++, vdcConfiguration.c_str(), false);
}

vector<int> DsVdcHost::readInstanceNumbers()
{
  vector<string> strs;
  vector<int> ints;

  boost::split(strs, vdcConfiguration, boost::is_any_of(";"));

  for (vector<string>::iterator strIt = strs.begin(); strIt != strs.end(); ++strIt) {
    if (!strIt->empty()) {
      ints.push_back(atoi(strIt->c_str()));
    }
  }

  // make sure that the instance numbers are sorted
  sort(ints.begin(), ints.end());

  return ints;
}

void DsVdcHost::saveInstanceNumbers(const vector<int>& aInstanceNumbers)
{
  vdcConfiguration = "";

  for (int i=0; i < aInstanceNumbers.size(); i++) {
    vdcConfiguration += string_format("%d;", aInstanceNumbers[i]);
  }
}

void DsVdcHost::addNewVdc()
{
  // add new instance to current list
  vdcCount++;
  vector<int> instanceNrs = readInstanceNumbers();
  instanceNrs.push_back(vdcCount);

  // save current configuration to db and schedule application reset
  saveInstanceNumbers(instanceNrs);
  this->markDirty();
  this->save();
  terminateTicket.executeOnce(boost::bind(&MainLoop::terminate, &MainLoop::currentMainLoop(), EXIT_SUCCESS), 5*Second);
}

void DsVdcHost::deleteVdc(VdcPtr aVdcPtr)
{
  if (!aVdcPtr) return;

  // we are trying to remove a specific vdc
  int instance = aVdcPtr->getInstanceNumber();

  // remove instance from current list
  vector<int> instanceNrs = readInstanceNumbers();
  instanceNrs.erase(std::remove(instanceNrs.begin(), instanceNrs.end(), instance), instanceNrs.end());

  // save current configuration to internal variable
  saveInstanceNumbers(instanceNrs);

  // clear the current configuration from this vdc
  aVdcPtr->reportVanished();
  aVdcPtr->forget();

  // if we have removed last vdc, make sure to add a new unconfigured one
  if (instanceNrs.size() == 0) {
    addNewVdc();
  } else {
    // save current configuration to db and schedule application reset
    this->markDirty();
    this->save();
    terminateTicket.executeOnce(boost::bind(&MainLoop::terminate, &MainLoop::currentMainLoop(), EXIT_SUCCESS), 5*Second);
  }

}


} /* namespace p44 */
