/*
 *  Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of vdc-utils
 */

// File scope debugging options
// - Set ALWAYS_DEBUG to 1 to enable DBGLOG output even in non-DEBUG builds of this file
#define ALWAYS_DEBUG 0
// - set FOCUSLOGLEVEL to non-zero log level (usually, 5,6, or 7==LOG_DEBUG) to get focus (extensive logging) for this file
//   Note: must be before including "logger.hpp" (or anything that includes "logger.hpp")
#define FOCUSLOGLEVEL 7


#include "httpserver.hpp"

#if __cplusplus >= 201402L

#include "thirdparty/boost/request_parser.hpp"
#include "macaddress.hpp"
#include "makeintrusive.hpp"


using namespace p44;

string HttpServer::HTTP_RESPONSE_OK = "HTTP/1.1 200 OK\r\n"
                                        "Content-Length: 0\r\n"
                                        "\r\n";

HttpServer::HttpServer(MainLoop &aMainLoop, const string& aHostname, const string& aPort) :
  inherited(aMainLoop)
{
  LOG(LOG_INFO,
      "- Configuring server:\n"
      "- host_name: '%s'\n"
      "- port: '%s'"
      "- ip: '%s'",
      aHostname.c_str(),
      aPort.c_str(),
      ipv4ToString(ipv4Address()).c_str()
  );

  setConnectionParams(aHostname.c_str(), aPort.c_str(), SOCK_STREAM, AF_INET);
  setAllowNonlocalConnections(true);
}


HttpServer::~HttpServer()
{
  closeConnection();
}


ErrorPtr HttpServer::startServer()
{
  if (!connectable()) {
    return TextError::err("Invalid server configuration");
  }
  return inherited::startServer([&](auto...params){ return this->configApiConnectionHandler(params...); }, MAX_CONNECTIONS);
}


p44::SocketCommPtr HttpServer::configApiConnectionHandler(SocketCommPtr aServerSocketCommP)
{
  auto conn = vdcutils::make_intrusive<JsonComm>(MainLoop::currentMainLoop());
  conn->setReceiveHandler([=](ErrorPtr aError){
    string receivedData;
    conn->receiveIntoString(receivedData);
    this->rawRequestHandler(conn, aError, receivedData);
  });
  conn->setClearHandlersAtClose();
  connections.push_back(make_pair(conn, MainLoop::now()));
  timeoutRoutineTicket.executeOnce([&](auto...){ this->timeoutRoutine(); }, CONNECTION_TIMEOUT);
  return conn;
}

void HttpServer::rawRequestHandler(JsonCommPtr aJsonComm, ErrorPtr aError, string& aRawHttpReq)
{
  LOG(LOG_INFO,
      "- Raw Data received: '%s'\n"
      "- Request error: '%s'",
      aRawHttpReq.c_str(),
      aError ? aError->description().c_str() : "ok");

  http::server::request_parser requestParser;
  Request req;

  http::server::request_parser::result_type result;
  string::iterator bodyIt;
  std::tie(result, bodyIt) = requestParser.parse(
      req.head, aRawHttpReq.begin(), aRawHttpReq.end());

  copy(bodyIt, aRawHttpReq.end(), back_inserter(req.body));

  if (aJsonComm) {
    aJsonComm->sendRaw(HTTP_RESPONSE_OK);
    aJsonComm->closeAfterSend();
  }

  if (dataHandler) {
    dataHandler(req.body, aError);
  }

  if (jsonDataHandler) {
    jsonDataHandler(JsonObject::objFromText(req.body.c_str()), aError);
  }
}


void HttpServer::timeoutRoutine()
{
  for (auto it = connections.begin(); it != connections.end(); ++it ) {
    if ((MainLoop::now() - it->second) > CONNECTION_TIMEOUT) {
      if (it->first) it->first->closeConnection();
      it = connections.erase(it);
    }
  }
}



#endif // c++14

