/*
 *  Copyright (c) 2020 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of vdc-utils
 */
#pragma once


#include "vdc-utils/persistentstorage/icontainer.hpp"
#include "vdc-utils/persistentstorage/persistentstorage.hpp"


namespace vdcutils {
namespace persistentstorage {


class ContainerException : public vdcutils::Exception
{
public:
  using vdcutils::Exception::Exception;
};


class ContainerCreateException : public ContainerException
{
public:
  using ContainerException::ContainerException;
};


class ContainerNotAvailable : public ContainerException
{
public:
  using ContainerException::ContainerException;
};


class ContainerQueryException : public ContainerException
{
public:
  using ContainerException::ContainerException;
};


class ParamStore : public p44::ParamStore
{
public:
  using p44::ParamStore::ParamStore;
protected:
  virtual std::string dbSchemaUpgradeSQL(int aFromVersion, int &aToVersion);
};


template<class T, template <typename> class C = std::vector>
class Container : public IContainer<T, C>
{
public:
  /**
   * @brief Container ctor creates database to store objects
   * @param aPath Path to database file
   * @param aTableName Name of the table in which objects will be stored
   * @param aReset If set database will be cleared
   */
  Container(ParamStore& aParamStore, const std::string& aTableName);
  virtual ~Container() = default;
  virtual void store() override;
  virtual void load() override;
  virtual C<T>& data() override;

protected:
  C<T> mData;
  typename T::type mValue;
  ParamStore& mParamStore;
  vdcutils::PersistentStorageWithTableName<p44::PersistentParams> mPersistentStorage;
};


constexpr auto CONTAINER_SCHEMA_MIN_VERSION = 1;
constexpr auto CONTAINER_SCHEMA_VERSION = 1;


template<class T, template <typename> class C>
Container<T,C>::Container(ParamStore& aParamStore, const std::string& aTableName) :
  mParamStore(aParamStore),
  mPersistentStorage(aTableName, std::make_tuple(
    PersistentRecordRef<typename T::type>({"value", mValue})
  ), mParamStore)
{
  if (!mParamStore.isAvailable()) {
    throw ContainerCreateException("Param storage is not available");
  }
}


template<class T, template <typename> class C>
void Container<T, C>::store()
{
  if (!mParamStore.isAvailable()) {
    throw ContainerNotAvailable();
  }
  {
    auto err = mPersistentStorage.clearTable();
    if (!p44::Error::isOK(err)) {
      throw ContainerQueryException("Could not clear table");
    }
  }
  for (const auto& item : mData) {
    // You need to load first to have correct rowid
    mPersistentStorage.load(item.rowID().c_str());
    mValue = item.serialize();
    mPersistentStorage.save(item.rowID());
  }
}


template<class T, template <typename> class C>
void Container<T, C>::load()
{
  if (!mParamStore.isAvailable()) {
    throw ContainerNotAvailable();
  }
  auto query = mPersistentStorage.newLoadAllQuery(nullptr);
  if (!query) {
    throw ContainerQueryException("Could not create load all query");
  }
  mData.clear();
  for (auto iter = query->begin(); iter != query->end(); ++iter) {
    uint64_t flags;
    int index = 0;
    mPersistentStorage.loadFromRow(iter, index, &flags);
    T item;
    item.deserialize(mValue);
    mData.push_back(std::move(item));
  }
}


template<class T, template <typename> class C>
C<T>& Container<T, C>::data()
{
  return mData;
}


} // name
} // namespace persistentstorage
