/*
 *  Copyright (c) 2020 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of vdc-utils
 */
#pragma once


#include "p44utils/persistentparams.hpp"
#include "vdc-utils/exception.hpp"
#include "vdc-utils/persistentstorage/persistentstorage.hpp"


namespace vdcutils {
namespace persistentstorage {


template <class T, typename Enable = void>
class ISerializable;


template<class T>
using FieldExists = std::enable_if_t<
    not std::is_void_v<typename vdcutils::detail::FieldType<T>::type>
  >;


template <class T>
class ISerializable<T, FieldExists<T>>
{
public:
  virtual ~ISerializable() = default;
  virtual T serialize() const = 0;
  virtual void deserialize(const T& data) = 0;
  virtual std::string rowID() const = 0;
  using type = T;
};


/**
 * Interface for container storing objects in database.
 * It is enabled if T implements ISerializable and C is supported container.
 */
template<class T, template <typename> class C = std::vector>
class IContainer
{
public:
  virtual ~IContainer() = default;

  /**
   * @brief Store object in database
   * @param aObj Object ot be stored, object need to implement ISerializable interface
   * Object should be uniquely identifiable by key. When object is already existing in
   * databes it will be updated.
   */
  virtual void store() = 0;

  /**
   * @brief Loads objects from database
   * Loads objects from database. If obecjt already exists in data it will be updated.
   * If object exisiting in data is not existing in DB it is not going to be deleted.
   */
  virtual void load() = 0;

  /**
   * @brief Provides access to stored objects
   * @return Reference to internal voctor with data.
   * Returned vecotr might be used to manipulate data without necessity to store modificaiton in DB.
   * To store or load data you need to manually use store or load function.
   */
  virtual C<T>& data() = 0;
};


} // name
} // namespace persistentstorage
