/*
 *  Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of vdc-utils
 */
#pragma once

#if __cplusplus >= 201402L

#include <string>
#include "p44utils/persistentparams.hpp"
#include "p44utils/jsonobject.hpp"
#include "ipersistentstoragewithrowid.hpp"
#include "ipersistentstoragefactory.hpp"

namespace vdcutils {

namespace detail {


class Field
{
protected:
  std::string name;
  p44::FieldDefinition fieldDef;

  Field(const std::string& aName, int aDataTypeCode) :
    name(aName),
    fieldDef({name.c_str(), aDataTypeCode})
  {}

public:
  virtual ~Field() {}
  p44::FieldDefinition* getFieldDef() { return &fieldDef; }
  virtual void bindToStatement(sqlite3pp::statement& aStatement, int aIndex) = 0;
  virtual void loadFromRow(sqlite3pp::query::iterator& aRow, int aIndex) = 0;
};


class StringField : public Field
{
  using inherited = Field;
  std::string& data;

public:
  StringField(const std::string& aName, std::string& aData) : inherited(aName, SQLITE_TEXT), data(aData) {}

  void bindToStatement(sqlite3pp::statement& aStatement, int aIndex) override
  {
    aStatement.bind(aIndex++, data.c_str(), false);
  }

  void loadFromRow(sqlite3pp::query::iterator& aRow, int aIndex) override
  {
    data = p44::nonNullCStr(aRow->get<const char *>(aIndex));
  }
};

class BoolField : public Field
{
  using inherited = Field;
  bool& data;

public:

  BoolField(const std::string& aName, bool& aData) : inherited(aName, SQLITE_INTEGER), data(aData) {}

  void bindToStatement(sqlite3pp::statement& aStatement, int aIndex) override
  {
    aStatement.bind(aIndex, data);
  }

  void loadFromRow(sqlite3pp::query::iterator& aRow, int aIndex) override
  {
    data = aRow->get<bool>(aIndex);
  }
};


class StringListField : public Field
{
  using inherited = Field;
  StringList& data;

public:
  StringListField(const std::string& aName, StringList& aData) : inherited(aName, SQLITE_TEXT), data(aData) {}

  void bindToStatement(sqlite3pp::statement& aStatement, int aIndex) override
  {
    auto array = p44::JsonObject::newArray();
    for(auto& elem : data) {
      array->arrayAppend(p44::JsonObject::newString(elem.c_str()));
    }
    aStatement.bind(aIndex++, array->c_strValue(), false);
  }

  void loadFromRow(sqlite3pp::query::iterator& aRow, int aIndex) override
  {
    data.clear();

    auto readData = p44::nonNullCStr(aRow->get<const char *>(aIndex));
    auto json = p44::JsonObject::objFromText(readData);
    if (!json) {
      return;
    }

    for(int i = 0 ; i < json->arrayLength(); i++) {
      data.push_back(json->arrayGet(i)->stringValue());
    }
  }
};


template <typename T>
struct FieldType {

};

template<>
struct FieldType<std::string> {
  using type = StringField;
};

template<>
struct FieldType<bool> {
  using type = BoolField;
};

template<>
struct FieldType<StringList> {
  using type = StringListField;
};

template <typename T>
class FieldWithCallback: public FieldType<T>::type
{
  using inherited = typename FieldType<T>::type;
  using LoadCB_t = typename PersistentRecordCB<T>::LoadCB_t;
  using SaveCB_t = typename PersistentRecordCB<T>::SaveCB_t;

  T mValue;
  LoadCB_t mLoadCB;
  SaveCB_t mSaveCB;

  public:
  FieldWithCallback(const std::string& aName, const LoadCB_t& aLoadCB, const SaveCB_t& aSaveCB):
    inherited(aName, mValue),
    mLoadCB(aLoadCB),
    mSaveCB(aSaveCB)
   { }

  void bindToStatement(sqlite3pp::statement& aStatement, int aIndex) override
  {
    mValue = mSaveCB();
    inherited::bindToStatement(aStatement, aIndex);
  }

  void loadFromRow(sqlite3pp::query::iterator& aRow, int aIndex) override
  {
    inherited::loadFromRow(aRow, aIndex);
    mLoadCB(mValue);
  }


};


}

template <typename BaseClass>
class PersistentStorage : public BaseClass
{
  std::vector<std::unique_ptr<detail::Field>> fieldDefs;

  using inherited = BaseClass;

  size_t numAllFields() { return fieldDefs.size(); }

  template <typename T>
  void createField(const std::vector<PersistentRecordRef<T>>& aFields)
  {
    for(auto& elem : aFields) {
      fieldDefs.push_back(std::make_unique<typename detail::FieldType<T>::type>(elem.name, elem.value));
    }
  }

  template <typename T>
  void createField(const std::vector<PersistentRecordCB<T>>& aFields)
  {
    for(auto& elem : aFields) {
      fieldDefs.push_back(std::make_unique<detail::FieldWithCallback<T>>(elem.name, elem.loadCB, elem.saveCB));
    }
  }

  template <typename T>
  void createField(const PersistentRecordCB<T>& aField)
  {
    fieldDefs.push_back(std::make_unique<detail::FieldWithCallback<T>>(aField.name, aField.loadCB, aField.saveCB));
  }

  template<typename T>
  void createField(const PersistentRecordRef<T>& aField)
  {
    fieldDefs.push_back(std::make_unique<typename detail::FieldType<T>::type>(aField.name, aField.value));
  }


  template <class Tuple, std::size_t... Is>
  void add(const Tuple& aTuple, std::index_sequence<Is...>)
  {
    (createField(std::get<Is>(aTuple)), ...);
  }

public:

  template <typename ...Args, typename ...BaseArgs>
  PersistentStorage(const std::tuple<Args...>& aRecords, BaseArgs&&... aBaseArgs) :
    inherited(std::forward<BaseArgs>(aBaseArgs)...)
  {
    add(aRecords, std::index_sequence_for<Args...>());
  }

  void load(const string& aRowId)
  {
    inherited::loadFromStore(aRowId.c_str());
  }

  void save(const string& aRowId)
  {
    inherited::markDirty();
    inherited::saveToStore(aRowId.c_str(), true);
  }

  size_t numFieldDefs() override
  {
    return inherited::numFieldDefs() + numAllFields();
  }

  const p44::FieldDefinition* getFieldDef(size_t aIndex) override
  {
    if (aIndex<inherited::numFieldDefs())
      return inherited::getFieldDef(aIndex);
    aIndex -= inherited::numFieldDefs();
    if (aIndex<numAllFields())
      return fieldDefs[aIndex]->getFieldDef();
    return nullptr;
  }


  void loadFromRow(sqlite3pp::query::iterator& aRow, int& aIndex, uint64_t *aCommonFlagsP) override
  {
    inherited::loadFromRow(aRow, aIndex, aCommonFlagsP);

    for(size_t i = 0; i < numAllFields(); i++) {
      fieldDefs[i]->loadFromRow(aRow, aIndex++);
    }
  }

  void bindToStatement(sqlite3pp::statement& aStatement, int& aIndex, const char *aParentIdentifier, uint64_t aCommonFlags) override
  {
    inherited::bindToStatement(aStatement, aIndex, aParentIdentifier, aCommonFlags);

    for(size_t i = 0 ; i < numAllFields(); i++) {
      fieldDefs[i]->bindToStatement(aStatement, aIndex++);
    }
  }
};

template <typename BaseClass>
class PersistentStorageWithTableName : public PersistentStorage<BaseClass>
{
  using inherited = PersistentStorage<BaseClass>;
  std::string name;

public:

  template <typename ...Args, typename ...BaseArgs>
  PersistentStorageWithTableName(const std::string& aTableName, const std::tuple<Args...>& aRecords, BaseArgs&&... aBaseArgs) :
    name(aTableName),
    inherited(aRecords, std::forward<BaseArgs>(aBaseArgs)...) {}


  const char *tableName() override { return name.c_str(); };

  p44::ErrorPtr clearTable()
  {
    p44::ErrorPtr err;
    if (inherited::paramStore.executef("DELETE FROM %s", tableName()) != SQLITE_OK) {
      err = inherited::paramStore.error();
    }
    if (p44::Error::notOK(err)) {
      LOG(LOG_ERR, "deleteFromStore: table=%s, failed: %s", tableName(), err->text());
    }
    return err;
  }

};

template <typename BaseClass>
class PersistentStorageWithRowId : public PersistentStorageWithTableName<BaseClass>, public IPersistentStorageWithRowId
{
  std::string rowId;

  using inherited = PersistentStorageWithTableName<BaseClass>;

public:

  template <typename ...Args, typename ...BaseArgs>
  PersistentStorageWithRowId(const string& aRowId, const std::string& aTableName, const std::tuple<Args...>& aRecords, BaseArgs&&... aBaseArgs) :
    rowId(aRowId),
    inherited(aTableName, aRecords, std::forward<BaseArgs>(aBaseArgs)...) {}

  void load() override
  {
    inherited::load(rowId);
  }

  void save() override
  {
    inherited::save(rowId);
  }

};

} // namespace vdcutils

#endif

