/*
 *  Copyright (c) 2019 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of vdc-utils
 */

#pragma once

#include <type_traits>
#include "p44vdc/vdc_common/propertycontainer.hpp"
#include "makeintrusive.hpp"

namespace vdcutils {


  template <typename T>
  using ReadCB = std::function<T()>;


  template <typename T>
  using WriteCB = std::function<void(const T&)>;


  template <typename T>
  struct Property
  {
    string name;
    ReadCB<T> readCB;
    WriteCB<T> writeCB;
  };


  class PropertyBase
  {
    string name;
    p44::PropertyDescription desc;

  public:

    PropertyBase(const std::string& aName, p44::ApiValueType aType, size_t aId, intptr_t aKey) :
      name(aName),
      desc({ name.c_str(), static_cast<uint16_t>(aType), aId, aKey }) {}

    virtual ~PropertyBase() {}

    virtual bool read(p44::ApiValuePtr aPropValue) = 0;
    virtual bool write(p44::ApiValuePtr aPropValue) = 0;
    p44::PropertyDescription* getDescription() { return &desc; }
  };

  template<typename PropertyType>
  class PropertyTemplate : public PropertyBase
  {
    using inherited = PropertyBase;

    ReadCB<PropertyType> readCB;
    WriteCB<PropertyType> writeCB;

    static constexpr p44::ApiValueType deduceType(){
      static_assert(
        std::is_same<PropertyType, bool>::value ||
        std::is_same<PropertyType, int64_t>::value ||
        std::is_same<PropertyType, string>::value,
        "Not supported property type.");

      if (std::is_same<PropertyType, bool>::value) return p44::apivalue_bool;
      if (std::is_same<PropertyType, int64_t>::value) return p44::apivalue_int64;
      if (std::is_same<PropertyType, string>::value) return p44::apivalue_string;
      return p44::apivalue_null;
    }

    void writeValue(p44::ApiValuePtr aPropValue, bool aValue) { aPropValue->setBoolValue(aValue); }
    void writeValue(p44::ApiValuePtr aPropValue, int64_t aValue) { aPropValue->setInt64Value(aValue); }
    void writeValue(p44::ApiValuePtr aPropValue, string aValue) { aPropValue->setStringValue(aValue); }

    template< typename T = PropertyType >
    typename std::enable_if_t< std::is_same< T, bool >::value, bool >
    readValue(p44::ApiValuePtr aPropValue) {
      return aPropValue->boolValue();
    }

    template< typename T = PropertyType >
    typename std::enable_if_t< std::is_same< T, int64_t >::value, int64_t >
    readValue(p44::ApiValuePtr aPropValue) {
      return aPropValue->int64Value();
    }

    template< typename T = PropertyType >
    typename std::enable_if_t<std::is_same< T, string >::value, string >
    readValue(p44::ApiValuePtr aPropValue) {
      return aPropValue->stringValue();
    }

  public:
    PropertyTemplate(const string& aName, size_t aId, intptr_t aKey, ReadCB<PropertyType> aReadCB, WriteCB<PropertyType> aWriteCB) :
      inherited(aName.c_str(), deduceType(), aId, aKey),
      readCB(std::move(aReadCB)),
      writeCB(std::move(aWriteCB)){}

    bool read(p44::ApiValuePtr aPropValue)
    {
      if (!readCB) {
        return false;
      }
      writeValue(aPropValue, readCB());

      return true;
    }

    bool write(p44::ApiValuePtr aPropValue)
    {
      if (!writeCB) {
        return false;
      }
      writeCB(readValue(aPropValue));

      return true;
    }
  };

} // namespace vdcutils
