/*
 *  Copyright (c) 2019 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of vdc-utils
 */

#pragma once

#include <deque>
#include <memory>
#include <functional>
#include "p44utils/error.hpp"
#include "p44utils/logger.hpp"



namespace vdcutils {

  template<typename T, typename Q>
  struct QueueCallback {
    using type = std::function<void(Q)>;
  };

  template<typename T, typename Q>
  struct Task {
    using type = std::function<void(std::function<void(T)>)>;
  };

  template<typename T, typename Q>
  struct TaskCallback {
    using type = std::function<void(T)>;
  };


  template<typename T, typename Q>
  class TaskQueueBase : public std::enable_shared_from_this<TaskQueueBase<T, Q>>
  {
    using Task_t = typename Task<T, Q>::type;
    using TaskCb_t = typename TaskCallback<T, Q>::type;
    using Queue_t = std::deque<std::pair<Task_t, TaskCb_t>>;
    using QueueCallback_t = typename QueueCallback<T, Q>::type;

    Queue_t mQueue;
    QueueCallback_t mFinishedCb;

    std::shared_ptr<TaskQueueBase<T, Q>> getPtr() {
      return this->shared_from_this();
    }

    protected:
    TaskQueueBase(){ }
    virtual ~TaskQueueBase(){ }

    public:
    void addTask(Task_t&& aTask);
    void addTaskFront(Task_t&& aTask);
    void executeQueue(QueueCallback_t&& aFinishCallback);
    virtual Q queueFinished() = 0;
    virtual Q queueAborted() = 0;
    void executeTask();
    virtual bool hasTaskSuccedded(T&& aResult) = 0;
    void handleTaskResult(T&& aResult);
    void abortQueue();

  };

  template<typename T, typename Q>
  void TaskQueueBase<T,Q>::addTask(Task_t&& aTask)
  {
    mQueue.emplace_back(std::make_pair(
      std::move(aTask),
      [thisPtr = getPtr()](auto...aArgs){ thisPtr->handleTaskResult(std::move(aArgs...)); })
    );
  }


  template<typename T, typename Q>
  void TaskQueueBase<T,Q>::addTaskFront(Task_t&& aTask)
  {
    mQueue.emplace_front(std::make_pair(
      std::move(aTask),
      [thisPtr = getPtr()](auto...aArgs){ thisPtr->handleTaskResult(std::move(aArgs...)); })
    );
  }


  template<typename T, typename Q>
  void TaskQueueBase<T,Q>::executeQueue(QueueCallback_t&& aFinishCallback)
  {
    LOG(LOG_DEBUG, "[TaskQueue] Executing queue, queued tasks '%d'", mQueue.size());
    mFinishedCb = std::move(aFinishCallback);
    executeTask();
  }


  template<typename T, typename Q>
  void TaskQueueBase<T,Q>::executeTask()
  {
    if (mQueue.empty()) {
      mFinishedCb(queueFinished());
      return;
    }
    auto [task, callback] = mQueue.front();
    mQueue.pop_front();
    task(callback);
  }


  template<typename T, typename Q>
  void TaskQueueBase<T,Q>::handleTaskResult(T&& aResult)
  {
    if (hasTaskSuccedded(std::move(aResult))){
      executeTask();
    } else {
      abortQueue();
      mFinishedCb(queueAborted());
    }
  }

  template<typename T, typename Q>
  void TaskQueueBase<T,Q>::abortQueue()
  {
    LOG(LOG_DEBUG, "[TaskQueue] Aborting task queue, suspending tasks '%d'", mQueue.size());
    Queue_t eq {};
    mQueue.swap(eq);
    LOG(LOG_DEBUG, "[TaskQueue] Ereased task queue, pending tasks '%d'", mQueue.size());
  }


  template<typename T = p44::ErrorPtr>
  class TaskQueue : public TaskQueueBase<T, p44::ErrorPtr>
  {
    public:
    using TaskQueueBase<T, p44::ErrorPtr>::TaskQueueBase;

    virtual p44::ErrorPtr queueFinished() override
    {
      return p44::Error::ok();
    }

    virtual p44::ErrorPtr queueAborted() override
    {
      return p44::TextError::err("Queue Aborted");
    }

    virtual bool hasTaskSuccedded(T&& aResult) override
    {
      return p44::Error::isOK(aResult);
    }
  };

  template<>
  inline bool TaskQueue<bool>::hasTaskSuccedded(bool&& aResult)
  {
    return aResult;
  }


} // namespace vdcutils
