/*
 *  Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of vdc-utils
 */

#include "devicetestbase.hpp"
#include "sensorbehaviour.hpp"
#include "testutils.hpp"


using namespace ::testing;
using namespace vdcutils;
using namespace vdcunittest;
using namespace p44;


DeviceTest::DeviceTest() :
  vdc(0, &host, 0)
{

}


void DeviceTest::identifyDeviceWithCallback()
{
  ErrorPtr err;
  EXPECT_CALL(*this, statusCbMock(_)).WillOnce(::testing::SaveArg<0>(&err));

  auto ret = device->identifyDevice([&](auto aErr, auto aDevice) {
    this->statusCbMock(aErr);
  });
  EXPECT_TRUE(Error::isOK(err));
  EXPECT_EQ(ret, false);
}


void DeviceTest::identifyDeviceWithReturn()
{
  auto ret = device->identifyDevice({});
  EXPECT_EQ(ret, true);
}


p44::ErrorPtr DeviceTest::handleMethod(const string& aMethod, ApiValuePtr aParams) {
  auto conn = make_intrusive<VdcJsonApiConnection>();
  auto req = make_intrusive<VdcJsonApiRequest>(conn, "");

  return device->handleMethod(req, aMethod, aParams);
}


void DeviceTest::getContainer(const std::string &aContainer, ContainerTest aCb)
{
  auto params = apiValuePtrFromJsonText(R"({ ")" + aContainer + R"(":null })");
  device->accessProperty(access_read, params, 0, 0, [=](ApiValuePtr aResultObject, ErrorPtr aError) {
    aCb(aResultObject->get(aContainer.c_str()));
  });
}


size_t DeviceTest::getContainerSize(ApiValuePtr aContainer) {
  EXPECT_TRUE(aContainer);
  size_t size = 0;
  string key;
  ApiValuePtr apiValue;
  aContainer->resetKeyIteration();

  while(aContainer->nextKeyValue(key, apiValue)) {
    size++;
  }

  return size;
}


bool DeviceTest::isContainerEmpty(ApiValuePtr aContainer)
{
  return (getContainerSize(aContainer) == 0);
}


void DeviceTest::testContainer(const std::string& aContainerName, const std::string& aTestJsonKey)
{
  getContainer(aContainerName, [this, aTestJsonKey](auto aContainer){
    EXPECT_TRUE(aContainer);

    auto elements = jsonTest->get(aTestJsonKey.c_str());
    EXPECT_TRUE(elements);

    testExpectedValues(aContainer, elements);
    testUnexpectedValues(aContainer, elements);
  });
}


void DeviceTest::testContainerWithDescription(const std::string& aContainerName, const std::string& aTestJsonKey)
{
  getContainer(aContainerName, [this, aTestJsonKey](auto aContainer){

    auto elements = jsonTest->get(aTestJsonKey.c_str());
    auto expectedValues = getExpectedValues(elements, getContainerSize(aContainer));

    for (int i=0; i<expectedValues->arrayLength(); ++i) {
      auto expectedValue = expectedValues->arrayGet(i);
      auto expectedValueName = getName(expectedValue);
      LOG(LOG_NOTICE, "Expecting '%s' in '%s'", expectedValueName.c_str(), aTestJsonKey.c_str());
      auto deviceValue = getDeviceValue(aContainer, expectedValueName);
      testDescription(deviceValue, expectedValue);
    }

    testUnexpectedValues(aContainer, elements);
  });
}


void DeviceTest::compareKeyValueContainers(p44::ApiValuePtr aDeviceContainer, p44::JsonObjectPtr aTestJson)
{
  EXPECT_TRUE(aDeviceContainer != nullptr);
  EXPECT_TRUE(aTestJson != nullptr);

  std::string key;
  p44::JsonObjectPtr value;
  size_t size{0};
  aTestJson->resetKeyIteration();

  while (aTestJson->nextKeyValue(key, value)) {
    LOG(LOG_NOTICE, "Expect key '%s' value '%s'", key.c_str(), value->stringValue().c_str());
    auto deviceValue = aDeviceContainer->get(key);
    EXPECT_TRUE(deviceValue != nullptr);
    EXPECT_EQ(deviceValue->stringValue(), value->stringValue());
    size++;
  }

  EXPECT_EQ(size, getContainerSize(aDeviceContainer));
}



/*
 *  deviceEventDescriptions:{
 *    BabycoolingStarted:{ description:"Boil and cool down (Babyboil) started" },
 *    BoilingAborted:{ description:"Boiling aborted by button press" },
 *    BoilingAbortedAndKettleReleased:{ description:"Boiling aborted by kettle being removed" },
 *    BoilingFinished:{ description:"Boiling ended without keeping warm" }
 *  }
 */

void DeviceTest::testEvents()
{
  testContainer("deviceEventDescriptions", "events");
}

/*
 *  deviceActionDescriptions:{
 *    boilandcooldown:{
 *      description:"boil and then cool down",
 *      params:{
 *        ParamCoolingTargetTemperature:{
 *          default:40.000000,
 *          max:99.000000,
 *          min:1.000000,
 *          optional:true,
 *          resolution:1.000000,
 *          siunit:"celsius",
 *          symbol:"°C",
 *          type:"numeric"
 *        },
 *        ParamHeatingTargetTemperature:{
 *          default:100.000000,
 *          max:100.000000,
 *          min:1.000000,
 *          optional:true,
 *          resolution:1.000000,
 *          siunit:"celsius",
 *          symbol:"°C",
 *          type:"numeric"
 *        },
 *      }
 *    },
 *    stop:{
 *      description:"stop",
 *      params:{  }
 *    }
 *  }
 */

void DeviceTest::testActions()
{
  getContainer("deviceActionDescriptions", [this](auto aContainer){

    auto actions = jsonTest->get("actions");
    auto expectedActions = getExpectedValues(actions, getContainerSize(aContainer));

    for (int i=0; i<expectedActions->arrayLength(); ++i) {
      auto expectedAction = expectedActions->arrayGet(i);
      auto expectedActionName = getName(expectedAction);
      LOG(LOG_NOTICE, "Expecting action '%s'", expectedActionName.c_str());
      auto deviceAction = getDeviceValue(aContainer, expectedActionName);
      testActionParams(deviceAction, expectedAction);
    }

    testUnexpectedValues(aContainer, actions);
  });
}

void DeviceTest::testActionParams(ApiValuePtr aDeviceAction, JsonObjectPtr aExpectedAction)
{
  auto expectedParams = aExpectedAction->get("params");
  auto deviceActionParams = aDeviceAction->get("params");

  if (!expectedParams) {
    LOG(LOG_NOTICE, "Action params not defined. Skipping...");
    return;
  }

  auto expectedParamsSize = expectedParams->arrayLength();
  EXPECT_EQ(expectedParamsSize, getContainerSize(deviceActionParams));
  if (expectedParamsSize > 0) {
    for (int i=0; i<expectedParamsSize; ++i) {
      auto expectedParam = expectedParams->arrayGet(i);
      auto expectedParamName = getName(expectedParam);
      LOG(LOG_NOTICE, "Expecting parameter '%s'", expectedParamName.c_str());
      auto deviceParam = getDeviceValue(deviceActionParams, expectedParamName);
      testDescription(deviceParam, expectedParam);
    }
  } else {
    LOG(LOG_NOTICE, "Parameters array is empty. Skipping...");
  }
}

void DeviceTest::testDescription(ApiValuePtr aCurrentParamDesc, JsonObjectPtr aTestValue)
{
  auto expectedDescription = aTestValue->get("description");
  if (!expectedDescription) {
    LOG(LOG_NOTICE, "Expected description not defined. Skipping...");
    return;
  }

  string expectedElemName;
  JsonObjectPtr expectedElemValue;
  expectedDescription->resetKeyIteration();

  while (expectedDescription->nextKeyValue(expectedElemName, expectedElemValue)) {
    LOG(LOG_NOTICE, "Expecting description element'%s' set to '%s'", expectedElemName.c_str(), expectedElemValue->stringValue().c_str());
    auto deviceParam = getDeviceValue(aCurrentParamDesc, expectedElemName);
    EXPECT_EQ(deviceParam->stringValue(), expectedElemValue->stringValue());
  }
}


p44::JsonObjectPtr DeviceTest::getExpectedValues(JsonObjectPtr aTestValues, size_t aExpectedSize)
{
  EXPECT_TRUE(aTestValues);
  auto expectedValues = aTestValues->get("expected");
  EXPECT_TRUE(expectedValues);
  EXPECT_EQ(expectedValues->arrayLength(), aExpectedSize);

  return expectedValues;
}

string DeviceTest::getName(JsonObjectPtr aTestValue)
{
  EXPECT_TRUE(aTestValue);
  auto name = aTestValue->get("name");
  EXPECT_TRUE(name);

  return name->stringValue();
}

p44::ApiValuePtr DeviceTest::getDeviceValue(ApiValuePtr aContainer, const string& aValueName)
{
  EXPECT_TRUE(aContainer);
  auto value = aContainer->get(aValueName.c_str());
  EXPECT_TRUE(value);

  return value;
}


/*
 *  devicePropertyDescriptions:{
 *    defaultcooldowntemperature:{
 *      max:100.000000,
 *      min:0.000000,
 *      resolution:1.000000,
 *      siunit:"celsius",
 *      symbol:"°C",
 *      type:"numeric"
 *    }
 *  }
*/

void DeviceTest::testProperties()
{
  testContainerWithDescription("devicePropertyDescriptions", "properties");
}


void DeviceTest::testExpectedValues(ApiValuePtr aContainer, JsonObjectPtr aValues)
{
  auto expectedValues = getExpectedValues(aValues, getContainerSize(aContainer));

  for (int i=0; i<expectedValues->arrayLength(); ++i) {
    auto expectedValue = expectedValues->arrayGet(i);
    EXPECT_TRUE(expectedValue);
    LOG(LOG_NOTICE, "Expecting '%s'", expectedValue->stringValue().c_str());
    EXPECT_TRUE(aContainer->get(expectedValue->stringValue().c_str()));
  }
}


void DeviceTest::testUnexpectedValues(ApiValuePtr aContainer, JsonObjectPtr aValues)
{
  EXPECT_TRUE(aContainer);
  EXPECT_TRUE(aValues);

  auto unexpected = aValues->get("unexpected");

  if (unexpected && (unexpected->arrayLength() > 0)) {
    for (int i=0; i<unexpected->arrayLength(); ++i) {
      auto elem = unexpected->arrayGet(i);
      EXPECT_TRUE(elem);
      LOG(LOG_NOTICE, "Not expecting '%s'", elem->stringValue().c_str());
      EXPECT_FALSE(aContainer->get(elem->stringValue().c_str()));
    }
  } else {
    LOG(LOG_NOTICE, "Unexpected entries not defined. Skipping...");
  }
}

/*
 *  deviceStateDescriptions:{
 *    operation:{
 *      description:"Operation",
 *      updateInterval:0.000000,
 *      value:{
 *        type:"enumeration",
 *        values:{
 *          cooldown:null,
 *          disconnected:null,
 *          heating:null,
 *          keepwarm:null,
 *          ready:null,
 *          removed:null
 *        }
 *      }
 *    }
 *  }
 */

void DeviceTest::testStates()
{
  getContainer("deviceStateDescriptions", [this](auto aContainer){

    auto states = jsonTest->get("states");
    auto expectedStates = getExpectedValues(states, getContainerSize(aContainer));

    for (int i=0; i<expectedStates->arrayLength(); ++i) {
      auto state = expectedStates->arrayGet(i);
      auto stateName = getName(state);
      LOG(LOG_NOTICE, "Expecting state '%s'", stateName.c_str());
      auto deviceEnumValues = getDeviceStateEnumValues(aContainer, stateName.c_str());
      testExpectedValues(deviceEnumValues, state);
      testUnexpectedValues(deviceEnumValues, state);
    }

    testUnexpectedValues(aContainer, states);
  });
}


p44::ApiValuePtr DeviceTest::getDeviceStateEnumValues(ApiValuePtr aContainer, const std::string &aDeviceState)
{
  EXPECT_TRUE(aContainer);
  auto deviceStateDescription = getDeviceValue(aContainer, aDeviceState);
  auto value = getDeviceValue(deviceStateDescription, "value");
  auto values = getDeviceValue(value, "values");

  return values;
}


p44::ApiValuePtr DeviceTest::getDeviceOutputValues(ApiValuePtr aContainer, const std::string &aDeviceState)
{
  EXPECT_TRUE(aContainer);
  auto outputDescription = getDeviceValue(aContainer, aDeviceState);
  auto values = getDeviceValue(outputDescription, "values");

  return values;
}

/*
 *  scenes:{
 *    64:{
 *      channels:{  },
 *      command:"",
 *      dontCare:true,
 *      effect:2,
 *      effectParam:0,
 *      ignoreLocalPriority:true
 *    },
 *    68:{
 *      channels:{  },
 *      command:"stop",
 *      dontCare:false,
 *      effect:1,
 *      effectParam:0,
 *      ignoreLocalPriority:true
 *    }
 *  }
 */

void DeviceTest::testScenes()
{
  getContainer("scenes", [this](auto aContainer){
    EXPECT_TRUE(aContainer);
    auto sceneCommands = jsonTest->get("scenes");
    EXPECT_TRUE(sceneCommands);
    sceneCommands->resetKeyIteration();
    string scene;
    JsonObjectPtr expectedCommand;

    while (sceneCommands->nextKeyValue(scene, expectedCommand)) {
      LOG(LOG_NOTICE, "Expecting command '%s' for scene '%s'", expectedCommand->stringValue().c_str(), scene.c_str());
      auto sceneNumber = sceneNumberFromName(scene);
      EXPECT_TRUE(sceneNumber);
      auto sceneDesc = getDeviceValue(aContainer, to_string(*sceneNumber));
      auto cmd = getDeviceValue(sceneDesc, "command");
      EXPECT_EQ(cmd->stringValue(), expectedCommand->stringValue());
    }
  });
}


boost::optional<DsSceneNumber> DeviceTest::sceneNumberFromName(const string& aSceneName)
{
  if (aSceneName == "FIRE") return FIRE;
  else if (aSceneName == "PANIC") return PANIC;
  else if (aSceneName == "STANDBY") return STANDBY;
  else if (aSceneName == "DEEP_OFF") return DEEP_OFF;
  else if (aSceneName == "SLEEPING") return SLEEPING;

  LOG(LOG_NOTICE, "Unknown command '%s'", aSceneName.c_str());
  return boost::none;
}


void DeviceTest::testIdentificationData()
{
  auto identifyData = jsonTest->get("identify");
  EXPECT_TRUE(identifyData);

  if (auto gtin = identifyData->get("gtin")) {
    EXPECT_EQ(device->oemModelGUID(), gtin->stringValue());
  }
  if (auto dSUID = identifyData->get("dSUID")) {
    EXPECT_EQ(device->getDsUid().getString(), dSUID->stringValue());
  }
  if (auto hwGUID = identifyData->get("hwGUID")) {
    EXPECT_EQ(device->hardwareGUID(), hwGUID->stringValue());
  }
  if (auto hwModelGUID = identifyData->get("hwModelGUID")) {
    EXPECT_EQ(device->hardwareModelGUID(), hwModelGUID->stringValue());
  }
  if (auto name = identifyData->get("name")) {
    EXPECT_EQ(device->getName(), name->stringValue());
  }
  if (auto modelName = identifyData->get("modelName")) {
    EXPECT_EQ(device->modelName(), modelName->stringValue());
  }
  if (auto vendorName = identifyData->get("vendorName")) {
    EXPECT_EQ(device->vendorName(), vendorName->stringValue());
  }
}


/*
 *  sensorDescriptions:{
 *    0:{
 *      name:"Current temperature",
 *      type:"sensor",
 *      dsIndex:0,
 *      x-p44-behaviourType:"sensor",
 *      sensorType:1,
 *      sensorUsage:4,
 *      siunit:"celsius",
 *      symbol:"°C",
 *      min:1.000000,
 *      max:100.000000,
 *      resolution:1.000000,
 *      updateInterval:0.000000,
 *      aliveSignInterval:0.000000
 *    }
 *  }
 */


void DeviceTest::testSensors()
{
  getContainer("sensorDescriptions", [this](auto aContainer){

    auto sensors = jsonTest->get("sensors");
    auto expectedSensors = getExpectedValues(sensors, getContainerSize(aContainer));

    for (int i=0; i<expectedSensors->arrayLength(); ++i) {
      auto expectedSensor = expectedSensors->arrayGet(i);
      auto expectedSensorName = getName(expectedSensor);
      LOG(LOG_NOTICE, "Expecting sensor '%s'", expectedSensorName.c_str());
      auto deviceSensor = device->getSensor(p44::Device::by_id, expectedSensorName);
      EXPECT_TRUE(deviceSensor);
      auto sensorId = deviceSensor->getIndex();
      auto deviceSensorDescription = getDeviceValue(aContainer, to_string(sensorId));
      testDescription(deviceSensorDescription, expectedSensor);
    }

    testUnexpectedValues(aContainer, sensors);
  });
}

/*
 *  {
 *    channelDescriptions:{
 *      videoInputSource:{
 *        channelType:26,
 *        dsIndex:3,
 *        max:17.000000,
 *        min:0.000000,
 *        name:"video input source",
 *        resolution:1.000000,
 *        siunit:"none",
 *        symbol:"",
 *        values:{
 *          Analog_TV:1, Apps:16, Component_1:9,
 *          DVB_C:5, DVB_S:3, DVB_T:2,
 *          Display_Port:14, Freesat:4, HDMI_1:10,
 *          HDMI_2:11, HDMI_3:12, HDMI_4:13,
 *          My_Apps:17, SAT_IP:6, SD_USB:15,
 *          TV:0, Video_1:7, Video_2:8
 *        }
 *      },
 *      videoStation:{
 *        channelType:25,
 *        dsIndex:2,
 *        max:9999.000000,
 *        min:0.000000,
 *        name:"video station",
 *        resolution:1.000000,
 *        siunit:"none",
 *        symbol:""
 *      }
 *    }
 *  }
 */


void DeviceTest::testOutputs()
{
  getContainer("channelDescriptions", [this](auto aContainer){

    auto outputChannels = jsonTest->get("outputs");
    auto expectedOutputs = getExpectedValues(outputChannels, getContainerSize(aContainer));

    for (int i=0; i<expectedOutputs->arrayLength(); ++i) {
      auto expectedOutput = expectedOutputs->arrayGet(i);
      auto expectedOutputName = getName(expectedOutput);
      LOG(LOG_NOTICE, "Expecting output channel '%s'", expectedOutputName.c_str());
      auto output = device->getOutput();
      EXPECT_TRUE(output != nullptr);
      auto deviceOutputChannel = output->getChannelById(expectedOutputName);
      EXPECT_TRUE(deviceOutputChannel != nullptr);

      auto outputDescription = getDeviceValue(aContainer, std::to_string(deviceOutputChannel->getChannelType()));
      testDescription(outputDescription, expectedOutput);

      auto testValues = expectedOutput->get("values");
      if (testValues) {
        LOG(LOG_NOTICE, "Test output values");
        auto deviceOutputValues = getDeviceOutputValues(aContainer, std::to_string(deviceOutputChannel->getChannelType()));
        compareKeyValueContainers(deviceOutputValues, testValues);
      }
    }

    testUnexpectedValues(aContainer, expectedOutputs);
  });
}


