/*
 *  Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of vdc-utils
 */

#include "gtest/gtest.h"
#include "mocks/persistentstoragefactorymock.hpp"
#include "mocks/persistentstoragewithrowidmock.hpp"
#include "vdc-utils/persistentstorage/persistentstorage.hpp"
#include "vdchost.hpp"

namespace vdcunittest {

using namespace ::testing;


class StorageTest : public Test
{
  public:
    using ExpectedStringFields = std::list<std::pair<std::string, std::string>>;

    PersistentStorageFactoryMock storageFactoryMock;
    StrictMock<PersistentStorageWithRowIdMock>* storageMock = nullptr;

    StorageTest() {

    }

    void testStorageFactory(const std::string &aExpRowId, const std::string &aExpTableName);
    void testCreateStringField(ExpectedStringFields &aExpectedFields);
    void testCreateStringListField(std::list<std::string> &aExpectedFields);

};


template<typename T>
class VdcStorageTest : public T, public Test
{
  public:
    using inherited = T;
    using ExpectedFieldDefs = std::list<p44::FieldDefinition>;
    static constexpr auto VDC_BASE_FIELDS_NUM = 8;
    p44::VdcHost host;

    static std::string getFieldType(const p44::FieldDefinition& aFieldDef);
    void testFieldDefinitions(ExpectedFieldDefs aExpectedFields);

    template<typename... VdcArgs>
    explicit VdcStorageTest(VdcArgs&&... aVdcArgs):
    T(0, &host, 0, std::forward<VdcArgs>(aVdcArgs)...)
    {

    }

};


template<typename T>
std::string VdcStorageTest<T>::getFieldType(const p44::FieldDefinition& aFieldDef)
{
  switch (aFieldDef.dataTypeCode)
  {
    case SQLITE_INTEGER:  return "INTEGER";
    case SQLITE_FLOAT:    return "FLOAT";
    case SQLITE_TEXT:     return "TEXT";
    case SQLITE_BLOB:     return "BLOB";

     default:
      return "UNKNOWN";
  }
}


template<typename T>
void VdcStorageTest<T>::testFieldDefinitions(ExpectedFieldDefs aExpectedFields)
{
  auto fieldsNum = inherited::numFieldDefs() - VDC_BASE_FIELDS_NUM;
  EXPECT_EQ(fieldsNum, aExpectedFields.size());
  for (int i=VDC_BASE_FIELDS_NUM; i<inherited::numFieldDefs(); ++i) {
    auto fieldDef = inherited::getFieldDef(i);
    EXPECT_TRUE(fieldDef);
    LOG(LOG_NOTICE, "Looking for field '%s' (%s) in expected fields...", fieldDef->fieldName, getFieldType(*fieldDef).c_str());
    auto it = find_if(aExpectedFields.begin(), aExpectedFields.end(), [=](auto aField){
      return (string(aField.fieldName) == string(fieldDef->fieldName));
    });
    EXPECT_TRUE(it != aExpectedFields.end());
    EXPECT_EQ(fieldDef->dataTypeCode, it->dataTypeCode);
  }
}

} // namespace vdcunittest


