/*
 *  Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of vdc-utils
 */
#pragma once

#if __cplusplus >= 201402L

#include "p44vdc/vdc_common/singledevice.hpp"
#include "p44vdc/vdc_common/jsonvdcapi.hpp"
#include "makeintrusive.hpp"
#include <boost/optional.hpp>

namespace vdcutils {

template <typename T>
class TypedEnumValueDescriptor : public p44::EnumValueDescriptor
{
  using inherited = p44::EnumValueDescriptor;
  using UnderlyingType = std::underlying_type_t<T>;

  std::map<T, std::string> enumNames;

public:
  TypedEnumValueDescriptor(const string& aName, const std::map<T, std::string>& aEnumNames) :
    inherited(aName),
    enumNames(aEnumNames)
  {
    for (auto const& e : enumNames) addEnum(e.second, e.first);
  }

  template <typename ...Args>
  TypedEnumValueDescriptor(const string& aName, Args... aArgs) :
    inherited(aName)
  {
    using expand_type = int[];
    expand_type { (enumNames.insert(std::make_pair(aArgs.second, aArgs.first)), 0)... };
    for (auto const& e : enumNames) addEnum(e.second, e.first);
  }

  boost::optional<T> fromString(const string& aName)
  {
    auto it = find_if(enumNames.begin(), enumNames.end(),
      [=](auto e){ return e.second == aName; });

    if (it != enumNames.end()) {
      return it->first;
    }
    return boost::none;
  }

  boost::optional<std::string> toString(T aValue)
  {
    auto it = find_if(enumNames.begin(), enumNames.end(),
      [=](auto e){ return e.first == aValue; });

    if (it != enumNames.end()) {
      return it->second;
    }
    return boost::none;
  }


  bool setValue(T aValue)
  {
    return setInt32Value(static_cast<UnderlyingType>(aValue));
  }

  void setDefaultValue(T aValue)
  {
    setValue(aValue);
    setIsDefault(true);
  }

  bool isValid()
  {
    return hasValue;
  }

  T getValue()
  {
    p44::JsonApiValuePtr jsonValuePtr = make_intrusive<p44::JsonApiValue>(p44::JsonApiValue());
    inherited::getValue(jsonValuePtr, true);
    return static_cast<T>(jsonValuePtr->int32Value());
  }

  T getPreviousValue()
  {
    p44::JsonApiValuePtr jsonValuePtr = make_intrusive<p44::JsonApiValue>(p44::JsonApiValue());
    inherited::getValue(jsonValuePtr, true, true);
    return static_cast<T>(jsonValuePtr->int32Value());
  }

  void addEnum(const string& aName, T aValue)
  {
    enumNames.insert(make_pair(aValue, aName));
    inherited::addEnum(aName.c_str(), static_cast<UnderlyingType>(aValue));
  }



};

template <typename T>
using TypedEnumValueDescriptorPtr = boost::intrusive_ptr<TypedEnumValueDescriptor<T>>;

} // namespace vdcutils

#endif

