/*
 *  Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of vdc-utils
 */
#pragma once

#if __cplusplus >= 201402L

#include "p44utils/jsonobject.hpp"
#include "p44utils/operationqueue.hpp"

using namespace std;

namespace vdcutils {

  enum class HttpMethod {
    GET,
    PUT,
    POST
  };

  const char* toCStr(HttpMethod aMethod);

  // MARK: ===== HttpOperation

  /// A handler that should be executed to start auth action
  using AuthCallback = boost::function<void (p44::StatusCB aStatusCallback)>;

  template <typename T>
  struct CallbackType { };

  template <typename T>
  struct DataType { };


  template <typename Client>
  class WebOperation : public p44::Operation
  {
    using inherited = p44::Operation;
    using Callback = typename CallbackType<Client>::type;
    using DataType = typename DataType<Client>::type;

    protected:
    Client &httpClient;                 ///< reference to communication class
    HttpMethod method;                      ///< http method
    string urlPath;                     ///< path to be called
    DataType data;                      ///< data to be sent or recevied data
    DataType response;                  ///< data received
    bool completed;                     ///< completion flag
    p44::ErrorPtr error;                     ///< communication error
    Callback resultHandler;             ///< callback to be executed at the end of operation
    AuthCallback authHandler;               ///< callback to be executed when authorization failed
    bool streamingReq;
    uint retryLimit;

    virtual void sendRequest() = 0;
    // restart the operation so it can be executed again
    void restartOperation();
    void issueRequest(const char *aContentType = nullptr, bool aSaveHeaders = false, bool aStreamResult = false);
    bool isAuthPending(p44::ErrorPtr aError);
    virtual void processAnswer(const DataType& aResponse, p44::ErrorPtr aError);
    /// Derived operation methods
    virtual bool initiate();
    virtual bool hasCompleted();
    virtual p44::OperationPtr finalize();
    virtual void abortOperation(p44::ErrorPtr aError);
    void setRetryLimit(uint aRetryLimit) { retryLimit = aRetryLimit; }

  private:

    void issueRequest(const string& aData, const char *aContentType, bool aSaveHeaders, bool aStreamResult);
    void issueRequest(p44::JsonObjectPtr aData, const char *aContentType, bool aSaveHeaders, bool aStreamResult);

  public:
    WebOperation(
      Client &aHttpClient,
      HttpMethod aMethod,
      const string& aUrlPath,
      const DataType& aData,
      Callback aResultHandler,
      AuthCallback aAuthHandler = NULL) :
       httpClient(aHttpClient),
       method(aMethod),
       urlPath(aUrlPath),
       data(aData),
       resultHandler(aResultHandler),
       authHandler(aAuthHandler),
       completed(false),
       streamingReq(false),
       retryLimit(3){}

    virtual ~WebOperation() P44_OVERRIDE {};

  };

  template <typename T>
  void WebOperation<T>::restartOperation()
  {
    retryLimit--;
    LOG(LOG_DEBUG, "- Restarting operation (%d)", retryLimit);
    if (retryLimit == 0) {
      this->abortOperation(p44::TextError::err("Retry limit exceeded, aborting operation..."));
      return;
    }

    initiated = false;
    completed = false;
    aborted = false;

    error = {};
    response = {};
  }

  template <typename T>
  void WebOperation<T>::issueRequest(const char *aContentType, bool aSaveHeaders, bool aStreamResult)
  {
    issueRequest(data, aContentType, aSaveHeaders, aStreamResult);
  }


  template <typename T>
  bool WebOperation<T>::isAuthPending(p44::ErrorPtr aError)
  {
    if (authHandler && httpClient.isAuthError(aError)) {
      // if we got an authorization error try to fix it and restart the operation
      authHandler([=](p44::ErrorPtr aAuthError) {
        if (p44::Error::isOK(aAuthError)) {
          this->restartOperation();
        } else {
          LOG(LOG_INFO, "Authorization not successful, aborting request to '%s'", urlPath.c_str());
          this->abortOperation(aAuthError);
        }
      });

      return true;
    } else {
      return false;
    }
  }

  template <typename T>
  bool WebOperation<T>::initiate()
  {
    if (!canInitiate(p44::MainLoop::now()))
      return false;
    // send http request
    sendRequest();
    // mark operation as initiated
    return inherited::initiate();
  }

  template <typename T>
  bool WebOperation<T>::hasCompleted() { return completed; }

  template <typename T>
  p44::OperationPtr WebOperation<T>::finalize()
  {
    if (resultHandler) {
      resultHandler(response, error);
      resultHandler = {}; // call once only
    }
    httpClient.processOperations();
    return inherited::finalize();
  }

  template <typename T>
  void WebOperation<T>::abortOperation(p44::ErrorPtr aError)
  {
    if (!aborted) {
      if (!completed) {
        httpClient.getApi().cancelRequest();
      }
      if (resultHandler && aError) {
        resultHandler({}, aError);
        resultHandler = {}; // call once only
      }
    }
    inherited::abortOperation(aError);
  }

  template <typename T>
  void WebOperation<T>::issueRequest(const string& aData, const char *aContentType, bool aSaveHeaders, bool aStreamResult)
  {
    LOG(LOG_DEBUG, "- Http Request to '%s' with data: '%s'", urlPath.c_str(), aData.c_str());
    streamingReq = aStreamResult;
    httpClient.getApi().httpRequest(urlPath.c_str(), [this](const string& aResp, p44::ErrorPtr aError){ this->processAnswer(aResp, aError); }, toCStr(method), aData.c_str(), aContentType, -1, aSaveHeaders, aStreamResult);
  }

  template <typename T>
  void WebOperation<T>::issueRequest(p44::JsonObjectPtr aData, const char *aContentType, bool aSaveHeaders, bool aStreamResult)
  {
    LOG(LOG_DEBUG, "- Json Request to '%s' with data: '%s'", urlPath.c_str(), aData ? aData->json_c_str() : "<no json data>");
    httpClient.getApi().jsonRequest(urlPath.c_str(), [this](auto... aParams){ this->processAnswer(aParams...); }, toCStr(method), aData, aContentType, aSaveHeaders);
  }


} // namespace vdcutils

#endif

