# WebSocket client

`vdcutils::websocket::Client` - secure websocket client (wss). This class is intended to be an interface to websocket functionality.

```
  void connect(
    std::string_view aHost,
    std::string_view aPort,
    std::string_view aTarget,
    SSLContextDecorator&& aSSLContext,
    std::shared_ptr<IEventProducer> aEventMonitor,
    p44::StatusCB aConnectedCB);
```
`connect` - establish websocket connection:
 - `aHost` - host ip
 - `aPort` - wss port
 - `aTarget` - target url
 - `aSSLContext` - SSL context providing any required data, such like api key, token etc.
 - `aEventMonitor` - websocket event monitor
 - `aConnectedCB` - result callback

```
  void writeData(const std::string& aData, p44::StatusCB aStatusCB);
```
`writeData` - send string data through websocket

```
  void readData(ReadCB_t aReadCB);
```
`readData` - read data from websocket once


`vdcutils::websocket::SSLContextDecorator` - decorator to configure SSL context. Decorate request adding required header fields, define SSL verify mode, add certificate path.

```
  SSLContextDecorator(
    SSLRequestDecorator_t aSSLRequestDecorator=[](auto...){},
    SSLVerifyMode_t aMode=SSL_VERIFY_PEER,
    std::string_view aCertPath={});
```
`ctor`:
 - `aSSLRequestDecorator` - request decorator handler
 - `aMode` - SSL verify mode, default is SSL_VERIFY_PEER
 - `aCertPath` - certificate path, default is none

`vdcutils::websocket::Error` - websocket error class:
 - `Resolve` - cannot resolve host
 - `Connect` - cannot connect
 - `Handshake` - handshake error
 - `SSLHandshake` - SSL handshake error
 - `IncompleteMessage` - received incomplete message
 - `CannotClose` - cannot close session
 - `OperationAborted` - websocket operation aborted
 - `OperationRetryLimit` - websocket operation retry exceeded
 - `ThreadHandlerException` - thread handler exception
 - `BoostSystemError` - boost system error


### Example:
```
// make a shared pointer as websocket::Client inherits from enable_shared_from_this
auto mWebSocketClient = std::make_shared<websocket::Client>(p44::MainLoop::currentMainLoop());

// create SSL context, decorate request with required attributes like token, api-key etc
auto sslContext = vdcutils::websocket::SSLContextDecorator(
  [] (auto& aRequest) {
    aRequest.insert(boost::beast::http::field::sec_websocket_version, "13");
    aRequest.insert(boost::beast::http::field::sec_websocket_extensions, "permessage-deflate; client_max_window_bits");
    aRequest.insert(boost::beast::http::field::sec_websocket_protocol, "v1.api.protocol.xyz");
    aRequest.insert("X-Api-Key", "xyz-xyz-xyz-xyz-xyz");
  },
  SSL_VERIFY_PEER, "device-root.pem");

// provide event monitor and handle wss events, see details of WebSocket event monitor
auto eventMonitor = std::make_shared<vdcutils::websocket::EventMonitor>();

// move SSL context to client and perform connection procedure
mWebSocketClient->connect(
  "192.168.0.136",
  "12345",
  "/wss/api",
  std::move(sslContext),
  eventMonitor,
  [=](auto aError){
    if (!p44::Error::isOK(aError)) {
      LOG(LOG_ERR, "Cannot connect")
      return;
    }
    // connection established, read data continuously
    mWebSocketClient->pollData([](auto aError, auto aData){
      LOG(LOG_NOTICE, "Received data '%s'", aData.c_str());
    });
});

  ...

auto const cmd = R"(
                      {
                          "command": "subscribe",
                          "device_id": "qwerty"
                      }
                  )";
// send data
mWebSocketClient->writeData(cmd1, [=](auto aError){
  if (p44::Error::isOK(aError)) {
    LOG(LOG_DEBUG, "Cmd sent sucessfully");
  }
});


```

# WebSocket event monitor

`vdcutils::websocket::EventMonitor` - Websocket event handler. Synchronizes websocket threads with p44::MainLoop.
`vdcutils::websocket::IEventProducer` - Websocket event producer. Notify event consumer.
`vdcutils::websocket::IEventConsumer` - Websocket event consumer. Wait for ws event.

### Example:
```
// setup Event handler
auto eventMonitor = std::make_shared<vdcutils::websocket::EventMonitor>();
mEventHandler = std::make_unique<EventHandler>(eventMonitor);
mEventHandler->setConnectedHandler(aStatusCB);
mEventHandler->setDataHandler([&](auto...aParams){ gotData(aParams...); });

// Event handler implemenatation
EventHandler::EventHandler(std::shared_ptr<vdcutils::websocket::IEventConsumer> aEventMonitor) :
  mEventMonitor(aEventMonitor)
{
}


void EventHandler::requestThread(p44::ChildThreadWrapper &aThread)
{
  auto event = mEventMonitor->waitForEvent();
  aThread.signalParentThread(vdcutils::toUnderlyingType(event));
}


void EventHandler::requestThreadSignal(p44::ChildThreadWrapper &aChildThread, p44::ThreadSignals aSignalCode)
{
  LOG_BUF(LOG_DEBUG, "Received signal from child thread: %d %p", aSignalCode, &p44::MainLoop::currentMainLoop());
  auto event = static_cast<vdcutils::websocket::Event>(aSignalCode);

  switch (event) {
    case vdcutils::websocket::Event::ThreadCompleted:
      mWsEventThread.reset();
      // continue, wait for next event
      if (!mTerminated) waitForEvent();
      return;
    case vdcutils::websocket::Event::ThreadCancelled:
      return;
    case vdcutils::websocket::Event::DataReady:
      auto data = mEventMonitor->getData();
      // process data
      return;
    case vdcutils::websocket::Event::Connected:
      if (mConnectedHandler) mConnectedHandler(mEventMonitor->getConnectedError());
      mConnectedHandler = {};
      return;
    case vdcutils::websocket::Event::Terminated:
      mDataHandler = {};
      if (mTerminateCB) mTerminateCB(p44::Error::ok());
      return;

    default:
      return;
  }
}


void EventHandler::waitForEvent()
{
  mWsEventThread = p44::MainLoop::currentMainLoop().executeInThread(
  [&](p44::ChildThreadWrapper &aThread) {
    requestThread(aThread);
  },
  [&](p44::ChildThreadWrapper &aChildThread, p44::ThreadSignals aSignalCode) {
    requestThreadSignal(aChildThread, aSignalCode);
  });
}


void EventHandler::terminate(p44::StatusCB aTerminatedCB)
{
  mTerminateCB = aTerminatedCB;
  mTerminated = true;
  mEventMonitor->terminate();
```
