/*
 *  Copyright (c) 2020 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of vdc-utils
 */

#pragma once

#include <condition_variable>
#include <mutex>
#include <list>
#include <string>
#include <memory>
#include "vdc-utils/loggerext.hpp"
#include "p44utils/mainloop.hpp"
#include "p44utils/error.hpp"

#define LOG_BUF(lvl, ...) LOG_M(lvl, "Event Monitor", ##__VA_ARGS__)


namespace vdcutils::websocket {


/* ===== Event monitor ===== */


enum class Event
{
  ThreadCompleted = p44::threadSignalCompleted,
  ThreadCancelled = p44::threadSignalCancelled,
  Connected = p44::threadSignalUserSignal,
  WriteDataCompleted,
  DataReady,
  Waiting,
  Terminated
};


using DataList_t = std::list<std::pair<p44::ErrorPtr, std::string>>;

class IEventConsumer
{
  public:
  virtual ~IEventConsumer() = default;
  virtual Event waitForEvent() = 0;
  virtual DataList_t getData() = 0;
  virtual p44::ErrorPtr getConnectedError() = 0;
  virtual void terminate() = 0;
};

class IEventProducer
{
  public:
  virtual ~IEventProducer() = default;
  virtual void pushData(p44::ErrorPtr aError, const std::string& aData) = 0;
  virtual void connected(p44::ErrorPtr aError) = 0;
  virtual void writeCompleted() = 0;
  virtual void terminate() = 0;
};


class EventMonitor : public IEventConsumer,
                     public IEventProducer
{
  std::mutex mMutex;
  std::condition_variable mConditionVar;
  DataList_t mDataBuffer;
  p44::ErrorPtr mConnectedError;
  Event mEvent{Event::Waiting};

  virtual Event waitForEvent() override;
  virtual void pushData(p44::ErrorPtr aError, const std::string& aData) override;
  virtual void connected(p44::ErrorPtr aError) override;
  virtual void writeCompleted() override;
  virtual DataList_t getData() override;
  virtual p44::ErrorPtr getConnectedError() override;
  virtual void terminate() override;

  public:
  virtual ~EventMonitor()=default;
};


} // namespace vdcutils
