/*
 *  Copyright (c) 2019 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of vdc-utils
 */

#include "vdc-utils/websocket/session.hpp"
#include "vdc-utils/websocket/error.hpp"
#include "vdc-utils/loggerext.hpp"
#include <future>

using namespace vdcutils::websocket;


/* ===== SSL Context ===== */


SSLContextDecorator::SSLContextDecorator(SSLContextDecorator&& aSSLContextDecorator) :
  mContext(std::move(aSSLContextDecorator.mContext)),
  mRequestDecorator(aSSLContextDecorator.mRequestDecorator)
{

}


SSLContextDecorator::SSLContextDecorator(
  SSLContextDecorator::SSLRequestDecorator_t aSSLRequestDecorator/*=[](auto...){}*/,
  SSLContextDecorator::SSLVerifyMode_t aMode/*=SSL_VERIFY_PEER*/,
  std::string_view aCertPath/*={}*/
) :
  mContext(boost::asio::ssl::context::sslv23_client),
  mRequestDecorator(aSSLRequestDecorator)
{
  mContext.set_verify_mode(aMode);
  if (!aCertPath.empty()) {
    mContext.load_verify_file(std::string(aCertPath));
  }
}


/* ===== Session ===== */


Session::Session(
    std::string_view aHost,
    std::string_view aPort,
    std::string_view aTarget,
    SSLContextDecorator&& aSSLContext,
    std::shared_ptr<IEventProducer> aEventMonitor) :
  mHost(aHost),
  mPort(aPort),
  mTarget(aTarget),
  mSslContext(std::move(aSSLContext)),
  mResolver(mService.mIOService),
  mWebSocket(mService.mIOService, mSslContext.mContext),
  mEventMonitor(aEventMonitor)
{
}


Session::~Session()
{
}


void Session::close(p44::StatusCB aClosedCB)
{
  LOG_WSS(LOG_NOTICE, "Closing session");
  try {
    mWebSocket.async_close(
      CLOSE_REASON_NORMAL,
      [this, self = shared_from_this(), aClosedCB] (auto aErrorCode) {
        LOG_WSS(LOG_NOTICE, "Session closed");
        if (aErrorCode) {
          aClosedCB(Error::err(aErrorCode));
          return;
        }
        aClosedCB(p44::Error::ok());
      }
    );
  } catch (std::exception& e) {
    aClosedCB(Error::err(Error::ErrorCode::CannotClose));
    LOG_WSS(LOG_ERR, "Session close failed. Reason '%s'", e.what());
  }
}


void Session::open(p44::StatusCB aConnectedCB)
{
  mService.async([this, self = shared_from_this(), aConnectedCB] {
    resolve([this, aConnectedCB](auto aError){
      mEventMonitor->connected(aError);
      aConnectedCB(aError);
    });
  });
}


void Session::resolve(p44::StatusCB aStatusCB)
{
  mService.async([this, self = shared_from_this(), aStatusCB] {
    mResolver.async_resolve(
      mHost,
      mPort,
      [this, aStatusCB](auto... aParams){ resolved(aParams..., aStatusCB); }
    );
  });
}


void Session::resolved(boost::system::error_code aErrorCode, Resolver_t::results_type aResults, p44::StatusCB aStatusCB)
{
  if (aErrorCode) {
    LOG_WSS(LOG_ERR, "Cannot resolve host '%s' port '%s'. Reason '%s'",
      mHost.c_str(), mPort.c_str(), aErrorCode.message().c_str());
      aStatusCB(Error::err(Error::ErrorCode::Resolve));
    return;
  }

    LOG_WSS(LOG_NOTICE, "Host '%s' resolved successfully", mHost.c_str());
    connect(aResults, aStatusCB);
}


void Session::connect(const Resolver_t::results_type& aResults, p44::StatusCB aStatusCB)
{
  boost::asio::async_connect(
    getTCPSocket(),
    aResults.begin(), aResults.end(),
    [this, self = shared_from_this(), aStatusCB] (auto aErrorCode, auto...) {
      if (aErrorCode) {
        LOG_WSS(LOG_ERR, "Cannot connect to host '%s' port '%s'. Reason '%s'",
          mHost.c_str(), mPort.c_str(), aErrorCode.message().c_str());
        aStatusCB(Error::err(Error::ErrorCode::Connect));
        return;
      }

      LOG_WSS(LOG_NOTICE, "Connected succesfully");
      handshake(aStatusCB);
    }
  );
}


void Session::handshake(p44::StatusCB aStatusCB)
{
  getSSLStream().async_handshake(
      SSL_HANDSHAKE_TYPE_CLIENT,
      [this, self = shared_from_this(), aStatusCB] (auto aErrorCode) {
        if (aErrorCode) {
          LOG_WSS(LOG_ERR, "Handshake failed. Reason '%s'", aErrorCode.message().c_str());
          aStatusCB(Error::err(Error::ErrorCode::Handshake));
          return;
        }
        LOG_WSS(LOG_NOTICE, "Handshake succesfull");
        sslHandshake(aStatusCB);
      }
  );
}


void Session::sslHandshake(p44::StatusCB aStatusCB)
{
  mWebSocket.async_handshake_ex(
    mHost,
    mTarget,
    mSslContext.mRequestDecorator,
    [this, self = shared_from_this(), aStatusCB] (auto aErrorCode) {
      if (aErrorCode) {
        LOG_WSS(LOG_ERR, "SSL handshake failed. Reason '%s'", aErrorCode.message().c_str());
        aStatusCB(Error::err(Error::ErrorCode::SSLHandshake));
        return;
      }
      LOG_WSS(LOG_NOTICE, "SSL Handshake succesfull");
      aStatusCB(p44::Error::ok());
    }
  );
}


void Session::readData(p44::StatusCB aStatusCB)
{
  mService.async(
    [this, self = shared_from_this(), aStatusCB]{
      mWebSocket.async_read(
        mDataBuffer,
        [this, aStatusCB] (auto aErrorCode, auto...) {
          if (aErrorCode) {
            LOG_WSS(LOG_ERR, "Read error. Reason '%s'", aErrorCode.message().c_str());
            aStatusCB(Error::err(aErrorCode));
            return;
          }

          if (!mWebSocket.is_message_done()) {
            LOG_WSS(LOG_ERR, "Incomplete message")
            aStatusCB(Error::err(Error::ErrorCode::IncompleteMessage));
            return;
          }

          auto dataStr = boost::beast::buffers_to_string(mDataBuffer.data());
          mDataBuffer.consume(dataStr.size());
          mEventMonitor->pushData(p44::Error::ok(), dataStr);
          aStatusCB(p44::Error::ok());
        }
      );
    }
  );
}


void Session::writeData(const std::string& aData, p44::StatusCB aStatusCB)
{
  if (!mWebSocket.is_open()) {
    LOG_WSS(LOG_ERR, "Cannot write. WebSocket is not connected");
    if (aStatusCB) (Error::err(Error::ErrorCode::NotConnected));
    return;
  }

  mService.async(
    [this, self = shared_from_this(), aData, aStatusCB] {
      mWebSocket.async_write(
        boost::asio::buffer(aData),
        [this, aStatusCB] (auto aErrorCode, auto...) {
          if (aErrorCode) {
            LOG_WSS(LOG_ERR, "Write error. Reason '%s'", aErrorCode.message().c_str());
            aStatusCB(Error::err(aErrorCode));
            return;
          }
          LOG_WSS(LOG_NOTICE, "Write success");
          aStatusCB(p44::Error::ok());
        }
      );
    }
  );
}
