/*
 *  Copyright (c) 2019 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of vdc-utils
 */

#pragma once

#include <string_view>
#include <functional>
#include <boost/beast/core.hpp>
#include <boost/beast/websocket.hpp>
#include <boost/beast/websocket/ssl.hpp>
#include <boost/asio/connect.hpp>
#include <boost/asio/ip/tcp.hpp>
#include <boost/asio/ssl/stream.hpp>

#include "p44utils/mainloop.hpp"
#include "vdc-utils/websocket/ioservice.hpp"
#include "vdc-utils/websocket/eventmonitor.hpp"

#define LOG_WSS(lvl, ...) LOG_M(lvl, "WebSocket Session", ##__VA_ARGS__)

namespace vdcutils::websocket {


/* ===== SSL Context ===== */


struct SSLContextDecorator
{
  using SSLRequestDecorator_t = std::function<void(boost::beast::websocket::request_type&)>;
  using SSLVerifyMode_t = boost::asio::ssl::context_base::verify_mode;

  boost::asio::ssl::context mContext;
  SSLRequestDecorator_t mRequestDecorator;

  SSLContextDecorator(
    SSLRequestDecorator_t aSSLRequestDecorator=[](auto...){},
    SSLVerifyMode_t aMode=SSL_VERIFY_PEER,
    std::string_view aCertPath={});
  SSLContextDecorator(SSLContextDecorator&& aSSLContextDecorator);
};

using ReadCB_t = std::function<void(p44::ErrorPtr, const std::string&)>;



/* ===== Websocket session ===== */


class Session : public std::enable_shared_from_this<Session>
{
  using Resolver_t = boost::asio::ip::tcp::resolver;
  using TCPSocket_t = boost::asio::ip::tcp::socket;
  using SSLStream_t = boost::asio::ssl::stream<TCPSocket_t>;
  using WebSocket_t = boost::beast::websocket::stream<SSLStream_t>;
  using Buffer_t = boost::beast::multi_buffer;
  static constexpr auto SSL_HANDSHAKE_TYPE_CLIENT = boost::asio::ssl::stream_base::client;
  static constexpr auto CLOSE_REASON_NORMAL = boost::beast::websocket::close_code::normal;

  const std::string mHost;
  const std::string mPort;
  const std::string mTarget;

  IOServiceHanlder mService;
  SSLContextDecorator mSslContext;
  Resolver_t mResolver;
  WebSocket_t mWebSocket;
  Buffer_t mDataBuffer;
  std::shared_ptr<IEventProducer> mEventMonitor;

  public:
  Session(
    std::string_view aHost,
    std::string_view aPort,
    std::string_view aTarget,
    SSLContextDecorator&& aSSLContext,
    std::shared_ptr<IEventProducer> aEventMonitor);
  ~Session();

  void open(p44::StatusCB aConnectedCB);
  void writeData(const std::string& aData, p44::StatusCB aStatusCB);
  void readData(p44::StatusCB aStatusCB);
  void close(p44::StatusCB aClosedCB);
  bool isOpen() const { return mWebSocket.is_open(); }

  private:
  void openThreadRoutine(p44::ChildThreadWrapper& aThread, p44::StatusCB aStatusCB);
  void resolve(p44::StatusCB aStatusCB);
  void resolved(boost::system::error_code aErrorCode, Resolver_t::results_type aResults, p44::StatusCB aStatusCB);
  void connect(const Resolver_t::results_type& aResults, p44::StatusCB aStatusCB);
  void handshake(p44::StatusCB aStatusCB);
  void sslHandshake(p44::StatusCB aStatusCB);

  TCPSocket_t& getTCPSocket() { return mWebSocket.next_layer().next_layer(); }
  SSLStream_t& getSSLStream() { return mWebSocket.next_layer(); }
};


} // namespace vdcutils
