/*
 *  Copyright (c) 2017 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of ds-vdc
 */

// File scope debugging options
// - Set ALWAYS_DEBUG to 1 to enable DBGLOG output even in non-DEBUG builds of this file
#define ALWAYS_DEBUG 0
// - set FOCUSLOGLEVEL to non-zero log level (usually, 5,6, or 7==LOG_DEBUG) to get focus (extensive logging) for this file
//   Note: must be before including "logger.hpp" (or anything that includes "logger.hpp")
#define FOCUSLOGLEVEL 7

#include "xmlobject.hpp"

#if __cplusplus >= 201402L


#include "p44utils/logger.hpp"
#include "p44utils/utils.hpp"
#include <fstream>
#include <sstream>


using namespace std;
using namespace rapidxml;
using namespace vdcutils;

XmlDocument::XmlDocument(const string& aTextXml)
{
  unique_ptr<char[]> tcPtr (new char [aTextXml.size() + 1]);
  textContent = move(tcPtr);
  copy(aTextXml.begin(), aTextXml.end() + 1, textContent.get());
  xmlObject.parse<0>(textContent.get());

}

XmlDocumentPtr XmlDocument::parseXmlText(const std::string& aXmlText)
{
  try {
    return std::make_shared<XmlDocument>(aXmlText);
  } catch (rapidxml::parse_error& e) {
    LOG(LOG_ERR, "XmlObject::XmlObject XML parse error: %s", e.what());
    return nullptr;
  }


}


XmlDocument::~XmlDocument()
{
  xmlObject.clear();
}


std::string XmlDocument::toString(const rapidxml::xml_document<>& aXmlDoc)
{
  std::string xmlString;
  rapidxml::print(std::back_inserter(xmlString), aXmlDoc, 0);

  return xmlString;
}



bool XmlNode::boolValue() const
{
  auto value = node.value();

  if(value == nullptr) {
    return false;
  }

  if (p44::lowerCase(value) == p44::lowerCase("true")) {
    return true;
  }

  return false;
}

string XmlNode::stringValue() const
{
  auto value = node.value();

  if(value == nullptr) {
    return "";
  }
  return string(value);
}

int32_t XmlNode::int32Value() const
{
  auto value = node.value();

  if(value == nullptr) {
    return 0;
  }
  return atoi(value);
}

double XmlNode::doubleValue() const
{
  auto value = node.value();

  if(value == nullptr) {
    return 0;
  }
  return atof(value);
}

bool XmlNode::setValue(const std::string& aValue)
{
  auto doc = node.document();
  if (!doc) return false;
  auto value = doc->allocate_string(aValue.c_str());
  node.value(value);
  return true;
}

XmlNodePtr XmlNode::get(const std::string& aKey) const
{
  auto child = node.first_node(aKey.c_str());

  if (child == nullptr) {
    return XmlNodePtr();
  }
  return std::make_shared<XmlNode>(*child, xml);
}

xml_attribute<> * XmlNode::getAttribute(const std::string& aName)
{
  return node.first_attribute(aName.c_str());
}

std::string XmlNode::name()
{
  return string(node.name());
}

XmlNodePtr XmlNode::getParent()
{
  auto parent = node.parent();

  if (parent == nullptr) {
    return XmlNodePtr();
  }
  return std::make_shared<XmlNode>(*parent, xml);
}

XmlNodePtr XmlNode::getFirstChild() const
{
  auto child = node.first_node();
  if (child == nullptr) {
    return XmlNodePtr();
  }
  return std::make_shared<XmlNode>(*child, xml);
}

XmlNodePtr XmlNode::nextSibling() const
{
  auto sibling = node.next_sibling();
  if (sibling == nullptr) {
    return XmlNodePtr();
  }
  return std::make_shared<XmlNode>(*sibling, xml);
}

XmlNodePtr XmlNode::nextSibling(const std::string& aNode)
{
  auto sibling = node.next_sibling(aNode.c_str());
  if (sibling == nullptr) {
    return XmlNodePtr();
  }
  return std::make_shared<XmlNode>(*sibling, xml);
}

XmlNodePtr XmlNode::findSubNode(const std::string& aNode)
{
  if (strlen(node.name()) > 0) {
    // if we have found the node, return it
    if ((strcmp(aNode.c_str(), node.name()) == 0) || aNode.empty()) {
      return std::make_shared<XmlNode>(node, xml);
    }

    // if this is not the node, search all of its subnodes
    for (auto child = getFirstChild(); child; child = child->nextSibling()) {
      auto foundNode = child->findSubNode(aNode);
      if (foundNode) {
        return foundNode;
      }
    }
  }
  return XmlNodePtr();
}

std::list<XmlNodePtr> XmlNode::findAllSubNodes(const std::string& aNode) const
{
  list<XmlNodePtr> nodeList;

  if (strlen(node.name()) > 0) {
    // if we have found the node, add it to the output
    if ((strcmp(aNode.c_str(), node.name()) == 0) || aNode.empty()) {
      nodeList.push_back(std::make_shared<XmlNode>(node, xml));
    }

    // gather the matching nodes from all subnodes
    for (auto n = getFirstChild(); n; n=n->nextSibling()) {
      auto nodes = n->findAllSubNodes(aNode);
      std::copy(nodes.begin(), nodes.end(), std::back_inserter(nodeList));
    }
  }

  return nodeList;
}

std::optional<bool> XmlNode::getBoolElem(const std::string& aNode) const
{
  std::optional<bool> retValue;
  auto child = node.first_node(aNode.c_str());

  if (child == nullptr) {
    return std::nullopt;
  }
  auto value = child->value();

  if (value == nullptr) {
    return std::nullopt;
  }
  if (p44::lowerCase(value) == p44::lowerCase("true")) {
    return true;
  }
  if (p44::lowerCase(value) == p44::lowerCase("false")) {
    return false;
  }
  return std::nullopt;
}

std::optional<double> XmlNode::getDoubleElem(const string& aNode) const
{
  auto child = node.first_node(aNode.c_str());

  if (child == nullptr) {
    return std::nullopt;
  }
  auto value = child->value();

  if (value == nullptr) {
    return std::nullopt;
  }
  return atof(value);
}

std::optional<int> XmlNode::getIntElem(const std::string& aNode) const
{
  auto child = node.first_node(aNode.c_str());

  if (child == nullptr) {
    return std::nullopt;
  }
  auto value = child->value();

  if (value == nullptr) {
    return std::nullopt;
  }
  return atoi(value);
}

std::optional<string> XmlNode::getStringElem(const std::string& aNode) const
{
  auto child = node.first_node(aNode.c_str());

  if (child == nullptr) {
    return std::nullopt;
  }
  auto value = child->value();

  if (value == nullptr) {
    return std::nullopt;
  }
  return string(value);
}


XmlObject::XmlObject(const string& aTextXml) : xml(std::make_shared<XmlDocument>(aTextXml))
{
}

std::optional<XmlObject> XmlObject::fromFile(const std::string& aPath)
{
  std::ifstream xmlFile(aPath);
  if (!xmlFile.is_open())
  {
    return std::nullopt;
  }
  std::stringstream xmlBase;
  xmlBase << xmlFile.rdbuf();
  xmlFile.close();
  try {
    return XmlObject(xmlBase.str());
  } catch (rapidxml::parse_error& e) {
    LOG(LOG_ERR, "e: %s", e.what());
    return std::nullopt;
  }
}

const xml_document<> * XmlObject::getXmlObject() const
{
  auto& doc = xml->getDocument();
  if (doc.first_node()) {
    return &doc;
  }

  return nullptr;
}

XmlNodePtr XmlObject::get(const std::string& aKey) const
{
  if (!getXmlObject()) {
    return XmlNodePtr();
  }
  auto elem = getXmlObject()->first_node()->first_node(aKey.c_str());

  if (!elem) {
    return XmlNodePtr();
  }
  return std::make_shared<XmlNode>(*elem, xml);
}

XmlNodePtr XmlObject::getRoot() const
{
  if (!getXmlObject()) {
    return XmlNodePtr();
  }
  auto elem = getXmlObject()->first_node();
  return std::make_shared<XmlNode>(*elem, xml);
}

std::string XmlObject::getPath(rapidxml::xml_node<> *aNode)
{
  string path;

  while (aNode) {
    path = string("/") + string(aNode->name()) + path;
    aNode = aNode->parent();
  }

  return path;
}

XmlNodePtr XmlObject::findNode(const string& aNode) const
{
  if (getXmlObject() == nullptr) {
    return XmlNodePtr();
  }

  for (auto node = std::make_shared<XmlNode>(*getXmlObject()->first_node(), xml); node; node = node->nextSibling()) {
    if (aNode == node->name()) {
      return node;
    }

    auto subNode = node->findSubNode(aNode);
    if (subNode) {
      return subNode;
    }
  }
  return XmlNodePtr();
}

list<XmlNodePtr> XmlObject::findAllNodes(const string& aNode) const
{
  list<XmlNodePtr> nodeList;
  auto node = findNode(aNode);

  if (!node) {
    return nodeList;
  }

  for (auto n = node; n; n=n->nextSibling(aNode)) {
    nodeList.push_back(n);
  }

  return nodeList;
}

std::optional<string> XmlObject::getStringElem(const string& aNode) const
{
  auto firstNode = getXmlObject() ? getXmlObject()->first_node() : nullptr;
  if (firstNode) {
    XmlNode xmlNode(*firstNode, xml);
    return xmlNode.getStringElem(aNode);
  }
  return std::nullopt;
}

std::optional<int> XmlObject::getIntElem(const string& aNode) const
{
  auto firstNode = getXmlObject() ? getXmlObject()->first_node() : nullptr;
  if (firstNode) {
    XmlNode xmlNode(*firstNode, xml);
    return xmlNode.getIntElem(aNode);
  }
  return std::nullopt;
}

std::optional<double> XmlObject::getDoubleElem(const string& aNode) const
{
  auto firstNode = getXmlObject() ? getXmlObject()->first_node() : nullptr;
  if (firstNode) {
    XmlNode xmlNode(*firstNode, xml);
    return xmlNode.getDoubleElem(aNode);
  }
  return std::nullopt;
}

std::optional<bool> XmlObject::getBoolElem(const std::string& aNode) const
{
  auto firstNode = getXmlObject() ? getXmlObject()->first_node() : nullptr;
  if (firstNode) {
    XmlNode xmlNode(*firstNode, xml);
    return xmlNode.getBoolElem(aNode);
  }
  return std::nullopt;
}

std::optional<string> XmlObject::toString() const
{
  auto doc = getXmlObject();
  if (!doc) {
    return std::nullopt;
  }
  return XmlDocument::toString(*doc);
}

#endif

