/*
 *  Copyright (c) 2017 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of ds-vdc
 */
#pragma once

#if __cplusplus >= 201402L

#include "thirdparty/rapidxml/rapidxml.hpp"
#include "thirdparty/rapidxml/rapidxml_print.hpp"
#include <memory>
#include <list>
#include <optional>

namespace vdcutils {

  class XmlDocument;
  using XmlDocumentPtr = std::shared_ptr<XmlDocument>;

  class XmlDocument
  {
    std::unique_ptr<char[]> textContent;      ///< text content of XML handled by rapidxml parser
    rapidxml::xml_document<> xmlObject;                 ///< parsed XML

  public:
    XmlDocument(const std::string& aTextXml);
    static XmlDocumentPtr parseXmlText(const std::string& aXmlText);
    ~XmlDocument();

    rapidxml::xml_document<>& getDocument() { return xmlObject; }
    static std::string toString(const rapidxml::xml_document<>& aXmlDoc);
  };


  class XmlNode;
  using XmlNodePtr = std::shared_ptr<XmlNode>;


  class XmlNode
  {
  protected:
    rapidxml::xml_node<>& node;
    XmlDocumentPtr xml;

  public:
    XmlNode(rapidxml::xml_node<>& aNode, XmlDocumentPtr aXml) :
      node(aNode),
      xml(aXml) {}

    bool boolValue() const;
    std::string stringValue() const;
    int32_t int32Value() const;
    double doubleValue() const;

    bool setValue(const std::string& aValue);

    XmlNodePtr get(const std::string& aKey) const;
    rapidxml::xml_attribute<> * getAttribute(const std::string& aName);

    std::string name();
    XmlNodePtr getParent();
    XmlNodePtr getFirstChild() const;
    XmlNodePtr nextSibling() const;
    XmlNodePtr nextSibling(const std::string& aNode);

    XmlNodePtr findSubNode(const std::string& aNode);
    std::list<XmlNodePtr> findAllSubNodes(const std::string& aNode) const;

    std::optional<bool> getBoolElem(const std::string& aElem) const;
    std::optional<double> getDoubleElem(const std::string& aNode) const;
    std::optional<int> getIntElem(const std::string& aNode) const;
    std::optional<std::string> getStringElem(const std::string& aNode) const;

    template <typename EnumType>
    std::optional<EnumType> getEnumElem(const std::string& aElem) const
    {
      auto elem = getIntElem(aElem);
      if (elem) {
        return std::optional<EnumType>(static_cast<EnumType>(*elem));
      }
      return std::nullopt;
    }

  };

  class XmlObject
  {

    protected:

      XmlDocumentPtr xml;

    public:
      XmlObject(const std::string& aTextXml);
      static std::optional<XmlObject> fromFile(const std::string& aPath);

      /// Get xmlObject
      /// @return xmlObject pointer, if parsing failed return NULL
      const rapidxml::xml_document<> * getXmlObject() const;

      XmlNodePtr get(const std::string& aKey) const;
      XmlNodePtr getRoot() const;

      /// Explore xmlObject(static.xml) to find desired node
      /// @param aNode name of node to be found
      XmlNodePtr findNode(const std::string& aNode) const;
      /// Find all nodes with given name
      /// @param aNode node name
      /// @return list of node pointers
      std::list<XmlNodePtr> findAllNodes(const std::string& aNode) const;

      // gets the full path of this node from the document root in /node1/node2 form
      static std::string getPath(rapidxml::xml_node<> *aNode);

      bool isValid() const { return (getXmlObject() != nullptr); }

      std::optional<std::string> getStringElem(const std::string& aElem) const;
      std::optional<int> getIntElem(const std::string& aElem) const;
      std::optional<double> getDoubleElem(const std::string& aElem) const;
      std::optional<bool> getBoolElem(const std::string& aElem) const;
      std::optional<std::string> toString() const;

  };
  using XmlObjectPtr = std::shared_ptr<XmlObject>;

} // namespace vdcutils

#endif

